"""This is a sample docstring."""

import uuid
import os
import pytest
import jsonpickle
from mocks import MockWorkerCompletion, MockWorkerSubmission
from mocks import mock_put

import beeflow.task_manager as tm
from beeflow.common.wf_data import Task
import beeflow


@pytest.fixture
def flask_client():
    """Client lets us run flask queries."""
    app = tm.flask_app
    app.config['TESTING'] = True
    db_path = f'/tmp/{uuid.uuid4().hex}.db'
    app.config['TESTING_DB_PATH'] = db_path
    client = app.test_client()
    yield client
    os.remove(db_path)


@pytest.fixture
@tm.connect_db
def database(db):
    """Fixture for connecting to the TM DB."""
    return db


def generate_tasks(n):
    """Generate n tasks for testing."""
    return [
        Task(f'task-{i}', base_command=['ls', '/'], hints=[], requirements=[],
             inputs=[], outputs=[], stdout=None, stderr=None,
             workflow_id=uuid.uuid4().hex)
        for i in range(n)
    ]


@pytest.mark.usefixtures('flask_client', 'mocker')
def test_submit_task(flask_client, mocker, database):  # noqa
    """Create a workflow and get the ID back."""
    # Generate a task
    tasks = generate_tasks(1)
    tasks_json = jsonpickle.encode(tasks)

    response = flask_client.post('/bee_tm/v1/task/submit/',
                                 json={'tasks': tasks_json})

    mocker.patch('beeflow.task_manager.worker',
                 new_callable=MockWorkerSubmission)

    # Patch the connection object for WFM communication
    mocker.patch('beeflow.common.connection.Connection.put', mock_put)
    beeflow.task_manager.process_queues()

    msg = response.get_json()['msg']
    status = response.status_code

    job_queue = list(database.job_queue)

    # We should only have a single job on the queue
    assert len(job_queue) == 1
    job = job_queue[0]
    assert job['task'] == tasks[0]
    assert job['job_id'] == 1
    assert job['job_state'] == 'RUNNING'

    assert status == 200
    assert msg == 'Tasks Added!'


@pytest.mark.usefixtures('flask_client', 'mocker')
def test_completed_task(flask_client, mocker, database): # noqa
    """Tests how the task manager processes a completed task."""
    # 42 is the sample task ID
    mocker.patch('beeflow.task_manager.worker',
                 new_callable=MockWorkerCompletion)
    # Patch the connection object for WFM communication
    mocker.patch('beeflow.common.connection.Connection.put', mock_put)

    # This should notice the job is complete and empty the job_queue
    beeflow.task_manager.process_queues()
    job_queue = list(database.job_queue)
    assert len(job_queue) == 0


@pytest.mark.usefixtures('flask_client', 'mocker')
def test_remove_task(flask_client, mocker, database):  # noqa
    """Test cancelling a workflow and removing tasks."""
    task1, task2, task3 = generate_tasks(3)
    # Add a few tasks
    database.job_queue.push(task=task1, job_id=1, job_state='RUNNING')
    database.job_queue.push(task=task2, job_id=2, job_state='PENDING')
    database.job_queue.push(task=task3, job_id=3, job_state='PENDING')

    mocker.patch('beeflow.task_manager.worker',
                 new_callable=MockWorkerCompletion)

    response = flask_client.delete('/bee_tm/v1/task/')

    msg = response.get_json()['msg']
    print(msg)
    status = response.status_code
    assert status == 200
    assert msg.count('CANCELLED') == 3



"""LSF worker for workload management.

Builds command for submitting batch job.
"""

import subprocess

from beeflow.common.worker.worker import Worker


class LSFWorker(Worker):
    """The Worker for systems where LSF is the Workload Manager."""

    def __init__(self, bee_workdir, **kwargs):
        """Create a new LSF Worker object."""
        super().__init__(bee_workdir, **kwargs)

        # Table of LSF states for translation to BEE states
        self.bee_states = {'PEND': 'PENDING',
                           'RUN': 'RUNNING',
                           'DONE': 'COMPLETED',
                           'QUIT': 'FAILED',
                           'PSUSP': 'PAUSED',
                           'USUSP': 'PAUSED',
                           'SSUSP': 'PAUSED'}
        # Check for extra runner options
        self.runner_opts = kwargs['runner_opts'] if 'runner_opts' in kwargs else ''

    def write_script(self, task):
        """Build task script; returns filename of script."""
        task_text = self.build_text(task)
        task_script = f'{self.task_save_path(task)}/{task.name}-{task.id}.sh'
        with open(task_script, 'w', encoding='UTF-8') as script_f:
            script_f.write(task_text)
            script_f.close()
        return task_script

    def query_job(self, job_id):
        """Query lsf for job status."""
        job_st = subprocess.check_output(['bjobs', '-aX', str(job_id), '-noheader'],
                                         stderr=subprocess.STDOUT)
        if 'not found' in str(job_st):
            raise Exception
        job_state = self.bee_states[job_st.decode().split()[2]]
        return job_state

    def submit_job(self, script):
        """Worker submits job-returns (job_id, job_state)."""
        job_st = subprocess.check_output(['bsub', script], stderr=subprocess.STDOUT)
        job_id = int(job_st.decode().split()[1][1:-1])
        job_state = self.query_job(job_id)
        return job_id, job_state

    def submit_task(self, task):
        """Worker builds & submits script."""
        task_script = self.write_script(task)
        job_id, job_state = self.submit_job(task_script)
        return job_id, job_state

    def query_task(self, job_id):
        """Worker queries job; returns job_state."""
        job_state = self.query_job(job_id)
        return job_state

    def cancel_task(self, job_id):
        """Worker cancels job, returns job_state."""
        subprocess.check_output(['bkill', str(job_id)], stderr=subprocess.STDOUT)
        job_state = "CANCELLED"
        return job_state



"""Workflow profiling code."""
import json
import time


class WorkflowProfiler:
    """Class for profiling a single workflow's execution."""

    def __init__(self, workflow_name, output_path):
        """Construct the workflow profiler class for a workflow."""
        self.workflow_name = workflow_name
        self.output_path = output_path
        self._state_changes = []
        self._scheduling_results = []

    def add_state_change(self, task, next_state):
        """Save a change of state for a task (at each task state change)."""
        self._state_changes.append({
            'task_id': task.id,
            'task_name': task.name,
            # State is not stored in the task object
            # 'previous_state': task.state,
            'next_state': next_state,
            'timestamp': int(time.time()),
        })

    def add_scheduling_results(self, tasks, resources, allocations):
        """Add scheduling results (given the set of available resources)."""
        self._scheduling_results.append({
            'tasks': tasks,
            'resources': resources,
            'allocations': dict(allocations),
            'timestamp': int(time.time()),
        })

    def save(self):
        """Save the workflow results (run on workflow completion)."""
        with open(self.output_path, 'w', encoding='utf-8') as fp:
            json.dump({
                'state_changes': self._state_changes,
                'scheduling_results': self._scheduling_results,
            }, fp=fp, indent=4)



#!/usr/bin/env python3

"""Functions for managing the BEE depency container and associated bind mounts."""

import os
import re
import sys
import time
import shutil
import signal
import subprocess

from beeflow.common import log as bee_logging
from beeflow.wf_manager.resources import wf_utils
from beeflow.common.config_driver import BeeConfig as bc


dep_log = bee_logging.setup(__name__)


class NoContainerRuntime(Exception):
    """An exception for no container runtime like charliecloud or singularity."""


def check_container_runtime():
    """Check if the container runtime is currently installed."""
    # Needs to support singuarity as well
    if shutil.which("ch-convert") is None or shutil.which("ch-run") is None:
        dep_log.error("ch-convert or ch-run not found. Charliecloud required"
                      " for neo4j container.")
        raise NoContainerRuntime('')


def make_dep_dir():
    """Make a new bee dependency container directory."""
    bee_workdir = wf_utils.get_bee_workdir()
    bee_dir = f'{bee_workdir}/deps'
    bee_dir_exists = os.path.isdir(bee_dir)
    if not bee_dir_exists:
        os.makedirs(bee_dir)


def get_dep_dir():
    """Return the dependency directory path."""
    bee_workdir = wf_utils.get_bee_workdir()
    bee_container_dir = f'{bee_workdir}/deps/'
    return bee_container_dir


def get_container_dir():
    """Return the depency container path."""
    container_name = 'dep_container'
    return get_dep_dir() + container_name


def check_container_dir():
    """Return true if the container directory exists."""
    container_dir = get_container_dir()
    container_dir_exists = os.path.isdir(container_dir)
    return container_dir_exists


def setup_gdb_mounts(mount_dir):
    """Set up mount directories for the graph database."""
    data_dir = mount_dir + '/data'
    logs_dir = mount_dir + '/logs'
    run_dir = mount_dir + '/run'
    certs_dir = mount_dir + '/certificates'

    data_dir = os.path.join(data_dir, "data")
    os.makedirs(data_dir, exist_ok=True)

    logs_dir = os.path.join(logs_dir, "logs")
    os.makedirs(logs_dir, exist_ok=True)

    mount_dir = os.path.join(run_dir, "run")
    os.makedirs(mount_dir, exist_ok=True)

    gdb_certs_path = os.path.join(certs_dir, "certificates")
    os.makedirs(gdb_certs_path, exist_ok=True)


def setup_gdb_configs(mount_dir, bolt_port, http_port, https_port):
    """Set up GDB configuration.

    This function needs to be run before each new invocation since the
    config file could have changed.
    """
    container_path = get_container_dir()
    confs_dir = os.path.join(mount_dir, "conf")
    os.makedirs(confs_dir, exist_ok=True)
    gdb_configfile = shutil.copyfile(container_path + "/var/lib/neo4j/conf/neo4j.conf",
                                     confs_dir + "/neo4j.conf")
    dep_log.debug(gdb_configfile)

    with open(gdb_configfile, "rt", encoding="utf8") as cfile:
        data = cfile.read()

    bolt_config = r'#(dbms.connector.bolt.listen_address=):[0-9]*'
    data = re.sub(bolt_config, rf'\1:{bolt_port}', data)
    http_config = r'#(dbms.connector.http.listen_address=):[0-9]*'
    data = re.sub(http_config, rf'\1:{http_port}', data)
    https_config = r'#(dbms.connector.https.listen_address=):[0-9]*'
    data = re.sub(https_config, rf'\1:{https_port}', data)
    with open(gdb_configfile, "wt", encoding="utf8") as cfile:
        cfile.write(data)


def create_image():
    """Create a new BEE dependency container if one does not exist.

    By default, the container is stored in /tmp/<user>/beeflow/deps.
    """
    # Can throw an exception that needs to be handled by the caller
    check_container_runtime()

    dep_img = bc.get('DEFAULT', 'bee_dep_image')

    # Check for BEE dependency container directory:
    container_dir_exists = check_container_dir()
    if container_dir_exists:
        dep_log.info('Already have neo4j container')
        return

    make_dep_dir()
    container_dir = get_container_dir()
    # Build new dependency container
    try:
        subprocess.run(["ch-convert", "-i", "tar", "-o", "dir",
                        str(dep_img), str(container_dir)],
                       stdout=sys.stdout, stderr=sys.stderr, check=True)
    except subprocess.CalledProcessError as error:
        dep_log.error(f"ch-convert failed: {error}")
        shutil.rmtree(container_dir)
        dep_log.debug(f"GraphDB container mount directory {container_dir} removed")
        return

    # Make the certificates directory
    container_certs_path = os.path.join(container_dir, 'var/lib/neo4j/certificates')
    os.makedirs(container_certs_path, exist_ok=True)


def start_gdb(mount_dir, bolt_port, http_port, https_port, reexecute=False):
    """Start the graph database."""
    setup_gdb_configs(mount_dir, bolt_port, http_port, https_port)
    # We need to rerun the mount step before each start
    if not reexecute:
        setup_gdb_mounts(mount_dir)

    db_password = bc.get('graphdb', 'dbpass')
    data_dir = mount_dir + '/data'
    logs_dir = mount_dir + '/logs'
    run_dir = mount_dir + '/run'
    certs_dir = mount_dir + '/certificates'
    confs_dir = mount_dir + "/conf"

    container_path = get_container_dir()
    if not reexecute:
        try:
            command = ['neo4j-admin', 'set-initial-password', str(db_password)]
            subprocess.run([
                "ch-run",
                "--set-env=" + container_path + "/ch/environment",
                "-b", confs_dir + ":/var/lib/neo4j/conf",
                "-b", data_dir + ":/data",
                "-b", logs_dir + ":/logs",
                "-b", run_dir + ":/var/lib/neo4j/run", container_path,
                "--", *command
            ], stdout=sys.stdout, stderr=sys.stderr, check=True)
        except subprocess.CalledProcessError:
            dep_log.error("neo4j-admin set-initial-password failed")
            return -1

    try:
        command = ['neo4j', 'start']
        with subprocess.Popen(["ch-run",
                              "--set-env=" + container_path + "/ch/environment",
                               "-b",
                               confs_dir + ":/var/lib/neo4j/conf",
                               "-b",
                               data_dir + ":/data",
                               "-b",
                               logs_dir + ":/logs",
                               "-b",
                               run_dir + ":/var/lib/neo4j/run",
                               "-b",
                               certs_dir + ":/var/lib/neo4j/certificates",
                               container_path,
                               "--", *command
                               ], stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
            output = proc.stdout.read().decode('utf-8')
            pid = re.search(r'pid ([0-9]*)', output).group(1)
            return pid
    except FileNotFoundError:
        dep_log.error("Neo4j failed to start.")
        return -1


def wait_gdb(log):
    """Need to wait for the GDB. Currently, we're using the sleep time paramater.

    We'd like to remove that in the future.
    """
    gdb_sleep_time = bc.get('graphdb', 'sleep_time')
    log.info(f'waiting {gdb_sleep_time}s for GDB to come up')
    time.sleep(gdb_sleep_time)


def remove_gdb():
    """Remove the current GDB bind mount directory."""
    bee_workdir = wf_utils.get_bee_workdir()
    gdb_workdir = os.path.join(bee_workdir, 'current_gdb')
    old_gdb_workdir = os.path.join(bee_workdir, 'old_gdb')
    if os.path.isdir(gdb_workdir):
        # Rename the directory to guard against NFS errors
        shutil.move(gdb_workdir, old_gdb_workdir)
        time.sleep(2)
        shutil.rmtree(old_gdb_workdir)
        time.sleep(2)


def kill_gdb(pid):
    """Kill the GDB with the associated pid."""
    try:
        os.kill(pid, signal.SIGTERM)
    except OSError:
        dep_log.info('Process already killed')



"""Logging interface for BEE."""
import logging
import os


# Set the default log level (BEE_LOG_LEVEL will be passed in by beeflow/cli.py)
LEVEL = os.getenv('BEE_LOG_LEVEL')
LEVEL = 'DEBUG' if LEVEL is None else LEVEL


def setup(name):
    """Set up and return logger.

    :param name: Name to be used for logger (best would be __name__)
    :type level: String
    """
    log = logging.getLogger(name)
    log.setLevel(LEVEL)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter('[%(asctime)s] %(name)s:%(funcName)s(): %(msg)s'))
    log.addHandler(handler)
    return log



"""Neo4j/Cypher transaction functions used by the Neo4jDriver class."""

from re import fullmatch


def create_workflow_node(tx, workflow):
    """Create a Workflow node in the Neo4j database.

    The workflow node is the entry point to the workflow.
    :param workflow: the workflow description
    :type workflow: Workflow
    """
    workflow_query = ("CREATE (w:Workflow) "
                      "SET w.id = $workflow_id "
                      "SET w.name = $name "
                      "SET w.state = $state")

    # Store the workflow ID and name in a new workflow node
    tx.run(workflow_query, workflow_id=workflow.id, name=workflow.name, state=workflow.state)


def create_workflow_hint_nodes(tx, hints):
    """Create Hint nodes for the workflow.

    :param hints: the workflow hints
    :type hints: list of Hint
    """
    for hint in hints:
        hint_query = ("MATCH (w:Workflow) "
                      "CREATE (w)<-[:HINT_OF]-(h:Hint $params) "
                      "SET h.class = $class_")

        tx.run(hint_query, params=hint.params, class_=hint.class_)


def create_workflow_requirement_nodes(tx, requirements):
    """Create Requirement nodes for the workflow.

    :param requirements: the workflow requirements
    :type requirements: list of Requirement
    """
    for req in requirements:
        req_query = ("MATCH (w:Workflow) "
                     "CREATE (w)<-[:REQUIREMENT_OF]-(r:Requirement $params) "
                     "SET r.class = $class_")

        tx.run(req_query, params=req.params, class_=req.class_)


def create_workflow_input_nodes(tx, inputs):
    """Create Input nodes for the workflow.

    :param inputs: the workflow inputs
    :type inputs: list of InputParameter
    """
    for input_ in inputs:
        input_query = ("MATCH (w:Workflow) CREATE (w)<-[:INPUT_OF]-(i:Input) "
                       "SET i.id = $input_id "
                       "SET i.type = $type "
                       "SET i.value = $value")

        tx.run(input_query, input_id=input_.id, type=input_.type, value=input_.value)


def create_workflow_output_nodes(tx, outputs):
    """Create Output nodes for the workflow.

    :param outputs: the workflow outputs
    :type outputs: list of OutputParameter
    """
    for output in outputs:
        output_query = ("MATCH (w:Workflow) CREATE (w)<-[:OUTPUT_OF]-(o:Output) "
                        "SET o.id = $output_id "
                        "SET o.type = $type "
                        "SET o.value = $value "
                        "SET o.source = $source")

        tx.run(output_query, output_id=output.id, type=output.type, value=output.value,
               source=output.source)


def create_task(tx, task):
    """Create a Task node in the Neo4j database.

    :param task: the new task to create
    :type task: Task
    """
    create_query = ("CREATE (t:Task) "
                    "SET t.id = $task_id "
                    "SET t.workflow_id = $workflow_id "
                    "SET t.name = $name "
                    "SET t.base_command = $base_command "
                    "SET t.stdout = $stdout "
                    "SET t.stderr = $stderr")

    # Unpack requirements, hints dictionaries into flat list
    tx.run(create_query, task_id=task.id, workflow_id=task.workflow_id, name=task.name,
           base_command=task.base_command, stdout=task.stdout, stderr=task.stderr)


def create_task_hint_nodes(tx, task):
    """Create Hint nodes for a task.

    :param task: the task whose hints to add to the graph
    :type task: Task
    """
    for hint in task.hints:
        hint_query = ("MATCH (t:Task {id: $task_id}) "
                      "CREATE (t)<-[:HINT_OF]-(h:Hint $params) "
                      "SET h.class = $class_")

        tx.run(hint_query, task_id=task.id, params=hint.params, class_=hint.class_)


def create_task_requirement_nodes(tx, task):
    """Create Requirement nodes for a task.

    :param task: the task whose requirements to add to the graph
    :type task: Task
    """
    for req in task.requirements:
        req_query = ("MATCH (t:Task {id: $task_id}) "
                     "CREATE (t)<-[:REQUIREMENT_OF]-(r:Requirement $params) "
                     "SET r.class = $class_")

        tx.run(req_query, task_id=task.id, params=req.params, class_=req.class_)


def create_task_input_nodes(tx, task):
    """Create Input nodes for a task.

    :param task: the task whose inputs to add to the graph
    :type task: Task
    """
    for input_ in task.inputs:
        input_query = ("MATCH (t:Task {id: $task_id}) "
                       "CREATE (t)<-[:INPUT_OF]-(i:Input) "
                       "SET i.id = $input_id "
                       "SET i.type = $type "
                       "SET i.value = $value "
                       "SET i.default = $default "
                       "SET i.source = $source "
                       "SET i.prefix = $prefix "
                       "SET i.position = $position "
                       "SET i.value_from = $value_from")

        tx.run(input_query, task_id=task.id, input_id=input_.id, type=input_.type,
               value=input_.value, default=input_.default, source=input_.source,
               prefix=input_.prefix, position=input_.position, value_from=input_.value_from)


def create_task_output_nodes(tx, task):
    """Create Output nodes for a task.

    :param task: the task whose outputs to add to the graph
    :type task: Task
    """
    for output in task.outputs:
        output_query = ("MATCH (t:Task {id: $task_id}) "
                        "CREATE (t)<-[:OUTPUT_OF]-(o:Output) "
                        "SET o.id = $output_id "
                        "SET o.type = $type "
                        "SET o.value = $value "
                        "SET o.glob = $glob")

        tx.run(output_query, task_id=task.id, output_id=output.id, type=output.type,
               value=output.value, glob=output.glob)


def create_task_metadata_node(tx, task):
    """Create a task metadata node in the Neo4j database.

    The node holds metadata about a task's execution state.

    :param task: the task for which to create a metadata node
    :type task: Task
    """
    metadata_query = ("MATCH (t:Task {id: $task_id}) "
                      "CREATE (m:Metadata {state: 'WAITING'})-[:DESCRIBES]->(t)")

    tx.run(metadata_query, task_id=task.id)


def add_dependencies(tx, task, old_task=None, restarted_task=False):
    """Create dependencies between tasks.

    :param task: the workflow task
    :type task: Task
    :param old_task: the failed task, ignored if not used with restarted_task=True
    :type old_task: Task
    :param restarted_task: restarted from failed task, only create dependencies for outputs
    :type restarted_task: bool
    """
    if restarted_task:
        delete_dependencies_query = ("MATCH (:Task {id: $task_id})<-[r:DEPENDS_ON]-(:Task) "
                                     "DETACH DELETE r")
        restarted_query = ("MATCH (s:Task {id: $old_task_id}), (t:Task {id: $new_task_id}) "
                           "MERGE (s)<-[:RESTARTED_FROM]-(t)")
        dependency_query = ("MATCH (s:Task {id: $task_id})<-[:OUTPUT_OF]-(o:Output) "
                            "WITH s, collect(o.id) AS outputs "
                            "MATCH (t:Task)<-[:INPUT_OF]-(i:Input) "
                            "WITH s, t, outputs, collect(i.source) as sources "
                            "WHERE any(output IN outputs WHERE output IN sources) "
                            "MERGE (t)-[:DEPENDS_ON]->(s)")

        tx.run(delete_dependencies_query, task_id=old_task.id)
        tx.run(restarted_query, old_task_id=old_task.id, new_task_id=task.id)
        tx.run(dependency_query, task_id=task.id)
    else:
        begins_query = ("MATCH (s:Task {id: $task_id})<-[:INPUT_OF]-(i:Input) "
                        "WITH s, collect(i.source) AS sources "
                        "MATCH (w:Workflow)<-[:INPUT_OF]-(i:Input) "
                        "WITH s, w, sources, collect(i.id) AS inputs "
                        "WHERE any(input IN sources WHERE input IN inputs) "
                        "MERGE (s)-[:BEGINS]->(w)")
        dependency_query = ("MATCH (s:Task {id: $task_id})<-[:INPUT_OF]-(i:Input) "
                            "WITH s, collect(i.source) as sources "
                            "MATCH (t:Task)<-[:OUTPUT_OF]-(o:Output) "
                            "WITH s, t, sources, collect(o.id) as outputs "
                            "WHERE any(input IN sources WHERE input IN outputs) "
                            "MERGE (s)-[:DEPENDS_ON]->(t) "
                            "WITH s "
                            "MATCH (s)<-[:OUTPUT_OF]-(o:Output) "
                            "WITH s, collect(o.id) AS outputs "
                            "MATCH (t:Task)<-[:INPUT_OF]-(i:Input) "
                            "WITH s, t, outputs, collect(i.source) as sources "
                            "WHERE any(output IN outputs WHERE output IN sources) "
                            "MERGE (t)-[:DEPENDS_ON]->(s)")

        tx.run(begins_query, task_id=task.id)
        tx.run(dependency_query, task_id=task.id)


def get_task_by_id(tx, task_id):
    """Get a workflow task from the Neo4j database by its ID.

    :param task_id: the task's ID
    :type task_id: str
    :rtype: BoltStatementResult
    """
    task_query = "MATCH (t:Task {id: $task_id}) RETURN t"

    return tx.run(task_query, task_id=task_id).single()


def get_task_hints(tx, task_id):
    """Get task hints from the Neo4j database by the task's ID.

    :param task_id: the task's ID
    :type task_id: str
    :rtype: BoltStatementResult
    """
    hints_query = "MATCH (:Task {id: $task_id})<-[:HINT_OF]-(h:Hint) RETURN h"

    return tx.run(hints_query, task_id=task_id)


def get_task_requirements(tx, task_id):
    """Get task requirements from the Neo4j database by the task's ID.

    :param task_id: the task's ID
    :type task_id: str
    :rtype: BoltStatementResult
    """
    reqs_query = "MATCH (:Task {id: $task_id})<-[:REQUIREMENT_OF]-(r:Requirement) RETURN r"

    return tx.run(reqs_query, task_id=task_id)


def get_task_inputs(tx, task_id):
    """Get task inputs from the Neo4j database by the task's ID.

    :param task_id: the task's ID
    :type task_id: str
    :rtype: BoltStatementResult
    """
    inputs_query = "MATCH (:Task {id: $task_id})<-[:INPUT_OF]-(i:Input) RETURN i"

    return tx.run(inputs_query, task_id=task_id)


def get_task_outputs(tx, task_id):
    """Get task outputs from the Neo4j database by the task's ID.

    :param task_id: the task's ID
    :type task_id: str
    :rtype: BoltStatementResult
    """
    outputs_query = "MATCH (:Task {id: $task_id})<-[:OUTPUT_OF]-(o:Output) RETURN o"

    return tx.run(outputs_query, task_id=task_id)


def get_workflow_description(tx):
    """Get the workflow description from the Neo4j database.

    :rtype: BoltStatementResult
    """
    workflow_desc_query = "MATCH (w:Workflow) RETURN w"

    return tx.run(workflow_desc_query).single()


def get_workflow_tasks(tx):
    """Get workflow tasks from the Neo4j database.

    :rtype: BoltStatementResult
    """
    workflow_query = "MATCH (t:Task) RETURN t"

    return tx.run(workflow_query)


def get_workflow_requirements(tx):
    """Get workflow requirements from the Neo4j database.

    :rtype: BoltStatementResult
    """
    requirements_query = "MATCH (:Workflow)<-[:REQUIREMENT_OF]-(r:Requirement) RETURN r"

    return tx.run(requirements_query)


def get_workflow_hints(tx):
    """Get workflow hints from the Neo4j database.

    :rtype: BoltStatementResult
    """
    hints_query = "MATCH (:Workflow)<-[:HINT_OF]-(h:Hint) RETURN h"

    return tx.run(hints_query)


def get_workflow_inputs(tx):
    """Get workflow inputs from the Neo4j database.

    :rtype: BoltStatementResult
    """
    inputs_query = "MATCH (:Workflow)<-[:INPUT_OF]-(i:Input) RETURN i"

    return tx.run(inputs_query)


def get_workflow_outputs(tx):
    """Get workflow outputs from the Neo4j database.

    :rtype: BoltStatementResult
    """
    outputs_query = "MATCH (:Workflow)<-[:OUTPUT_OF]-(o:Output) RETURN o"

    return tx.run(outputs_query)


def get_workflow_state(tx):
    """Get workflow state from the Neo4j database.

    :rtype: str
    """
    state_query = "MATCH (w:Workflow) RETURN w.state"

    return tx.run(state_query).single().value()


def set_workflow_state(tx, state):
    """Get workflow state from the Neo4j database.

    :param state: the state the workflow will be set to
    :type state: str
    """
    state_query = "MATCH (w:Workflow) SET w.state = $state"

    return tx.run(state_query, state=state)


def get_ready_tasks(tx):
    """Get all tasks that are ready to execute.

    :rtype: BoltStatementResult
    """
    get_ready_query = "MATCH (:Metadata {state: 'READY'})-[:DESCRIBES]->(t:Task) RETURN t"

    return tx.run(get_ready_query)


def get_dependent_tasks(tx, task):
    """Get the tasks that depend on a specified task.

    :param task: the task whose dependencies to obtain
    :type task: Task
    :rtype: BoltStatementResult
    """
    dependents_query = "MATCH (t:Task)-[:DEPENDS_ON]->(:Task {id: $task_id}) RETURN t"

    return tx.run(dependents_query, task_id=task.id)


def get_task_state(tx, task):
    """Get the state of a task.

    :param task: the task whose state to get
    :type task: Task
    :rtype: str
    """
    state_query = "MATCH (m:Metadata)-[:DESCRIBES]->(:Task {id: $task_id}) RETURN m.state"

    return tx.run(state_query, task_id=task.id).single().value()


def set_task_state(tx, task, state):
    """Set a task's state.

    :param task: the task whose state to set
    :type task: Task
    :param state: the new task state
    :type state: str
    """
    state_query = ("MATCH (m:Metadata)-[:DESCRIBES]->(:Task {id: $task_id}) "
                   "SET m.state = $state")

    tx.run(state_query, task_id=task.id, state=state)


def get_task_metadata(tx, task):
    """Get a task's metadata.

    :param task: the task whose metadata to get
    :type task: Task
    :rtype: BoltStatementResult
    """
    metadata_query = "MATCH (m:Metadata)-[:DESCRIBES]->(:Task {id: $task_id}) RETURN m"

    return tx.run(metadata_query, task_id=task.id).single()


def set_task_metadata(tx, task, metadata):
    """Set a task's metadata.

    :param task: the task whose metadata to set
    :type task: Task
    :param metadata: the task metadata
    :type metadata: dict
    """
    for k, v in metadata.items():
        # Manual sanitization needed for keys as the official Neo4j driver does not
        # currently support parameter substitution for property keys
        if fullmatch(r"[A-Za-z][0-9A-Za-z_]*", k) is None:
            raise ValueError(f"invalid metadata key: {k}")
        metadata_query = ("MATCH (m:Metadata)-[:DESCRIBES]->(:Task {id: $task_id}) "
                          f"SET m.{k} = $value")

        tx.run(metadata_query, task_id=task.id, value=v)


def get_task_input(tx, task, input_id):
    """Get a task input object.

    :param task: the task whose input to retrieve
    :type task: Task
    :param input_id: the ID of the input
    :type input_id: str
    :rtype: StepInput
    """
    input_query = ("MATCH (t:Task {id: $task_id})<-[:INPUT_OF]-(i:Input {id: $input_id}) "
                   "RETURN i")

    return tx.run(input_query, task_id=task.id, input_id=input_id).single()


def set_task_input(tx, task, input_id, value):
    """Set the value of a task input.

    :param task: the task whose input to set
    :type task: Task
    :param input_id: the ID of the input
    :type input_id: str
    :param value: str or int or float
    """
    input_query = ("MATCH (t:Task {id: $task_id})<-[:INPUT_OF]-(i:Input {id: $input_id}) "
                   "SET i.value = $value")

    tx.run(input_query, task_id=task.id, input_id=input_id, value=value)


def get_task_output(tx, task, output_id):
    """Get a task output object.

    :param task: the task whose output to retrieve
    :type task: Task
    :param output_id: the ID of the output
    :type output_id: str
    :rtype: StepOutput
    """
    output_query = ("MATCH (:Task {id: $task_id})<-[:OUTPUT_OF]-(o:Output {id: $output_id}) "
                    "RETURN o")

    return tx.run(output_query, task_id=task.id, output_id=output_id).single()


def set_task_output(tx, task, output_id, value):
    """Set a task's output value.

    :param task: the task whose output to set
    :type task: Task
    :param output_id: the ID of the output to set
    :type output_id: str
    :param value: the value of the output
    :type value: str
    """
    output_query = ("MATCH (:Task {id: $task_id})<-[:OUTPUT_OF]-(o:Output {id: $output_id}) "
                    "SET o.value = $value")

    tx.run(output_query, task_id=task.id, output_id=output_id, value=value)


def set_task_input_type(tx, task, input_id, type_):
    """Set the type of a task input.

    :param task: the task whose input type to set
    :type task: Task
    :param input_id: the ID of the input
    :type input_id: str
    :param type_: the input type to set
    :param type_: str
    """
    type_query = ("MATCH (:Task {id: $task_id})<-[:INPUT_OF]-(i:Input {id: $input_id}) "
                  "SET i.type = $type_")

    tx.run(type_query, task_id=task.id, input_id=input_id, type_=type_)


def set_task_output_glob(tx, task, output_id, glob):
    """Set a task's output value.

    :param task: the task whose output to set
    :type task: Task
    :param output_id: the ID of the output to set
    :type output_id: str
    :param glob: the glob of the output
    :type glob: str
    """
    glob_query = ("MATCH (:Task {id: $task_id})<-[:OUTPUT_OF]-(o:Output {id: $output_id}) "
                  "SET o.glob = $glob")

    tx.run(glob_query, task_id=task.id, output_id=output_id, glob=glob)


def set_init_tasks_to_ready(tx):
    """Set the initial workflow tasks' states to 'READY'."""
    init_ready_query = ("MATCH (m:Metadata)-[:DESCRIBES]->(t:Task)-[:BEGINS]->(:Workflow) "
                        "WHERE NOT (t)-[:DEPENDS_ON]->(:Task) "
                        "SET m.state = 'READY'")

    tx.run(init_ready_query)


def set_init_task_inputs(tx):
    """Set the initial workflow tasks' inputs from workfow inputs or defaults if necessary."""
    task_inputs_query = ("MATCH (i:Input)-[:INPUT_OF]->(:Task)-[:BEGINS]->(:Workflow)"
                         "<-[:INPUT_OF]-(wi:Input) "
                         "WHERE i.source = wi.id AND wi.value IS NOT NULL "
                         "SET i.value = wi.value")
    # Set any values to defaults if necessary
    defaults_query = ("MATCH (i:Input)-[:INPUT_OF]->(t:Task)-[:BEGINS]->(:Workflow)"
                      "<-[:INPUT_OF]-(wi:Input) "
                      "WHERE i.source = wi.id "
                      "AND i.value IS NULL AND i.default IS NOT NULL "
                      "SET i.value = i.default")

    tx.run(task_inputs_query)
    tx.run(defaults_query)


def copy_task_outputs(tx, task):
    """Use task outputs to set dependent task inputs or workflow outputs.

    Sets dependent task inputs to default value if necessary.

    :param task: the task whose outputs to set
    :type task: Task
    """
    task_inputs_query = ("MATCH (i:Input)-[:INPUT_OF]->(:Task)-[:DEPENDS_ON]->"
                         "(t:Task {id: $task_id})<-[:OUTPUT_OF]-(o:Output) "
                         "WHERE i.source = o.id AND o.value IS NOT NULL "
                         "SET i.value = o.value")
    # Set any values to defaults if necessary
    defaults_query = ("MATCH (m:Metadata)-[:DESCRIBES]->(:Task)<-[:DEPENDS_ON]-"
                      "(t:Task)-[:DEPENDS_ON]->(:Task {id: $task_id}) "
                      "WITH m, t "
                      "MATCH (t)<-[:INPUT_OF]-(i:Input) "
                      "WITH i, collect(m) AS mlist "
                      "WHERE all(m IN mlist WHERE m.state = 'COMPLETED') "
                      "AND i.value IS NULL AND i.default IS NOT NULL "
                      "SET i.value = i.default")
    workflow_output_query = ("MATCH (:Workflow)<-[:OUTPUT_OF]-(wo:Output) "
                             "WITH wo "
                             "MATCH (t:Task {id: $task_id})<-[:OUTPUT_OF]-(o:Output) "
                             "WHERE wo.source = o.id "
                             "SET wo.value = o.value")

    tx.run(task_inputs_query, task_id=task.id)
    tx.run(defaults_query, task_id=task.id)
    tx.run(workflow_output_query, task_id=task.id)


def set_running_tasks_to_paused(tx):
    """Set 'RUNNING' task states to 'PAUSED'."""
    set_paused_query = ("MATCH (m:Metadata {state: 'RUNNING'})-[:DESCRIBES]->(:Task) "
                        "SET m.state = 'PAUSED'")

    tx.run(set_paused_query)


def set_paused_tasks_to_running(tx):
    """Set 'PAUSED' task states to 'RUNNING'."""
    set_running_query = ("MATCH (m:Metadata {state: 'PAUSED'})-[:DESCRIBES]->(:Task) "
                         "SET m.state = 'RUNNING'")

    tx.run(set_running_query)


def set_runnable_tasks_to_ready(tx):
    """Set task states to 'READY' if all required inputs have values."""
    set_runnable_ready_query = ("MATCH (m:Metadata)-[:DESCRIBES]->"
                                "(t:Task)<-[:INPUT_OF]-(i:Input) "
                                "WITH m, t, collect(i) AS ilist "
                                "WHERE m.state = 'WAITING' "
                                "AND all(i IN ilist WHERE i.value IS NOT NULL) "
                                "SET m.state = 'READY'")

    tx.run(set_runnable_ready_query)


def reset_tasks_metadata(tx):
    """Reset the metadata for each of a workflow's tasks."""
    reset_metadata_query = ("MATCH (m:Metadata)-[:DESCRIBES]->(t:Task) "
                            "DETACH DELETE m "
                            "WITH t "
                            "CREATE (:Metadata {state: 'WAITING'})-[:DESCRIBES]->(t)")

    tx.run(reset_metadata_query)


def reset_workflow_id(tx, new_id):
    """Reset the workflow ID of the workflow using uuid4.

    :param new_id: the new workflow ID
    :type new_id: str
    """
    reset_workflow_id_query = ("MATCH (w:Workflow), (t:Task) "
                               "SET w.id = $new_id "
                               "SET t.workflow_id = $new_id")

    tx.run(reset_workflow_id_query, new_id=new_id)


def final_tasks_completed(tx):
    """Return true if each of a workflow's final Task nodes has state 'COMPLETED'.

    :rtype: bool
    """
    not_completed_query = ("MATCH (m:Metadata)-[:DESCRIBES]->(t:Task) "
                           "WHERE NOT (t)<-[:DEPENDS_ON|:RESTARTED_FROM]-(:Task) "
                           "AND m.state <> 'COMPLETED' "
                           "RETURN t IS NOT NULL LIMIT 1")

    # False if at least one task with state not 'COMPLETED'
    return bool(tx.run(not_completed_query).single() is None)


def is_empty(tx):
    """Return true if the database is empty, else false.

    :rtype: bool
    """
    empty_query = "MATCH (n) RETURN n IS NULL LIMIT 1"

    # False if at least one task
    return bool(tx.run(empty_query).single() is None)


def cleanup(tx):
    """Clean up all workflow data in the database."""
    cleanup_query = "MATCH (n) DETACH DELETE n"

    tx.run(cleanup_query)




"""Simple Worker class for launching tasks on a system with no workload manager."""

import subprocess

import os
from beeflow.common.worker.worker import Worker


class SimpleWorker(Worker):
    """Worker interface for system with no workload manager."""

    def __init__(self, container_runtime, **kwargs):
        """Create Simple worker object."""
        super().__init__(container_runtime=container_runtime, **kwargs)
        # This should be stored in Redis if possible
        self.tasks = {}

    def submit_task(self, task):
        """Worker submits task; returns job_id, job_state.

        :param task: instance of Task
        :rtype: tuple (int, string)
        """
        script = self.build_text(task)
        script_path = os.path.join(self.task_save_path(task), f'{task.name}-{task.id}.sh')
        with open(script_path, 'w', encoding='UTF-8') as fp:
            fp.write(script)
        with subprocess.Popen(['/bin/sh', script_path]) as taskid:
            self.tasks[task.id] = taskid
        return (task.id, 'PENDING')

    def cancel_task(self, job_id):
        """Cancel task with job_id; returns job_state.

        :param job_id: to be cancelled
        :type job_id: integer
        :rtype: string
        """
        self.tasks[job_id].kill()
        return 'CANCELLED'

    def query_task(self, job_id):
        """Query job state for the task.

        :param job_id: job id to query for status.
        :type job_id: int
        :rtype: string
        """
        return_code = self.tasks[job_id].poll()
        # This assumes a standard returncode
        if return_code is None:
            return 'RUNNING'
        if return_code == 0:
            return 'COMPLETED'
        return 'FAILED'



"""Task Manager database code."""
from contextlib import contextmanager
import sqlite3
import jsonpickle


class SubmitQueue:
    """Task Manager submit queue."""

    def __init__(self, conn):
        """Construct a submit queue handler."""
        self._conn = conn

    def __iter__(self):
        """Create an iterator for going over all elements."""
        cur = self._conn.cursor()
        res = cur.execute('SELECT task FROM submit_queue ORDER BY id ASC')
        while True:
            row = res.fetchone()
            if row is None:
                break
            task = jsonpickle.decode(row['task'])
            yield task

    def count(self):
        """Count the number of items in the submit queue."""
        res = self._conn.execute('SELECT COUNT(*) AS count FROM submit_queue').fetchone()
        return res['count']

    def push(self, task):
        """Push the task onto the submit queue."""
        cur = self._conn.cursor()
        task_data = jsonpickle.encode(task)
        cur.execute('INSERT INTO submit_queue (task) VALUES (:task)', {'task': task_data})
        self._conn.commit()

    def pop(self):
        """Pop the bottom element off the queue."""
        cur = self._conn.cursor()
        res = cur.execute('SELECT id, task FROM submit_queue ORDER BY id ASC').fetchone()
        id_ = res['id']
        print(id_)
        task_data = res['task']
        task = jsonpickle.decode(task_data)
        cur.execute('DELETE FROM submit_queue WHERE id = :id', {'id': id_})
        self._conn.commit()
        return task

    def clear(self):
        """Clear the submit queue."""
        self._conn.execute('DELETE FROM submit_queue')


class JobQueue:
    """Task Manager job queue."""

    def __init__(self, conn):
        """Construct a job queue handler."""
        self._conn = conn

    def __iter__(self):
        """Create an iterator for going over all elements in the queue."""
        res = self._conn.execute(
            'SELECT id, task, job_id, job_state FROM job_queue ORDER BY id ASC',
        )
        while True:
            row = res.fetchone()
            if row is None:
                break
            yield {
                'id': row['id'],
                'task': jsonpickle.decode(row['task']),
                'job_id': row['job_id'],
                'job_state': row['job_state'],
            }

    def count(self):
        """Count the number of items in the job queue."""
        res = self._conn.execute('SELECT COUNT(*) AS count FROM job_queue').fetchone()
        return res['count']

    def push(self, task, job_id, job_state):
        """Push the job info onto the queue."""
        cur = self._conn.cursor()
        task_data = jsonpickle.encode(task)
        cur.execute(
            """INSERT INTO job_queue (task, job_id, job_state)
               VALUES (:task, :job_id, :job_state)""",
            {'task': task_data, 'job_id': job_id, 'job_state': job_state},
        )
        self._conn.commit()

    def pop(self):
        """Pop the bottom element off the queue."""
        cur = self._conn.cursor()
        res = cur.execute(
            'SELECT id, task, job_id, job_state FROM job_queue ORDER BY id ASC',
        ).fetchone()
        id_ = res['id']
        task = jsonpickle.decode(res['task'])
        job_id = res['job_id']
        job_state = res['job_state']
        cur.execute('DELETE FROM job_queue WHERE id = :id', {'id': id_})
        self._conn.commit()
        return {'task': task, 'job_id': job_id, 'job_state': job_state}

    def update_job_state(self, id_, job_state):
        """Update the job_state."""
        cur = self._conn.cursor()
        cur.execute('UPDATE job_queue SET job_state = :job_state WHERE id = :id',
                    {'id': id_, 'job_state': job_state})
        self._conn.commit()

    def remove_by_id(self, id_):
        """Remove a job from the queue by ID."""
        cur = self._conn.cursor()
        cur.execute('DELETE FROM job_queue WHERE id = :id', {'id': id_})
        self._conn.commit()

    def clear(self):
        """Clear the job queue."""
        self._conn.execute('DELETE FROM job_queue')


class TMDB:
    """Task Manager Database."""

    def __init__(self, conn):
        """Construct a new TM database connection."""
        self._conn = conn
        self._init_tables()

    def _init_tables(self):
        """Initialize the workflow tables."""
        script = """CREATE TABLE IF NOT EXISTS submit_queue(
                        id INTEGER PRIMARY KEY ASC,
                        task TEXT,
                        popped INTEGER);
                    CREATE TABLE IF NOT EXISTS job_queue(
                        id INTEGER PRIMARY KEY ASC,
                        task TEXT,
                        job_id INTEGER,
                        job_state TEXT,
                        popped INTEGER);"""
        self._conn.executescript(script)

    @property
    def submit_queue(self):
        """Return a SubmitQueue object."""
        return SubmitQueue(self._conn)

    @property
    def job_queue(self):
        """Return a JobQueue object."""
        return JobQueue(self._conn)


@contextmanager
def open_db(fname):
    """Open and return a new database."""
    with sqlite3.connect(fname) as conn:
        conn.row_factory = sqlite3.Row
        yield TMDB(conn)



"""High-level BEE workflow management interface.

Delegates its work to a GraphDatabaseInterface instance.
"""

import re

from beeflow.common.wf_data import Workflow, Task


class WorkflowInterface:
    """Interface for manipulating workflows."""

    def __init__(self, gdb_interface):
        """Initialize the Workflow interface.

        Initializing this interface automatically attempts to
        connect to the graph database.

        :param gdb_interface: the gdb interface
        """
        # Connect to the graph database
        self._gdb_interface = gdb_interface
        # Store the Workflow ID in the interface to assign it to new task objects
        self._workflow_id = None

    def reconnect(self):
        """Reconnect to the graph database using stored credentials."""
        raise NotImplementedError()

    def initialize_workflow(self, workflow_id, name, inputs, outputs, requirements=None,
                            hints=None):
        """Begin construction of a BEE workflow.

        :param workflow_id: the pre-generated workflow ID
        :type workflow_id: str
        :param name: the workflow name
        :type name: str
        :param inputs: the inputs to the workflow
        :type inputs: list of InputParameter
        :param outputs: the outputs of the workflow
        :type outputs: list of OutputParameter
        :param requirements: the workflow requirements
        :type requirements: list of Requirement
        :param hints: the workflow hints (optional requirements)
        :type hints: list of Hint
        :rtype: Workflow
        """
        if self.workflow_loaded():
            raise RuntimeError("attempt to re-initialize existing workflow")
        if requirements is None:
            requirements = []
        if hints is None:
            hints = []

        workflow = Workflow(name, hints, requirements, inputs, outputs, workflow_id)
        self._workflow_id = workflow_id
        # Load the new workflow into the graph database
        self._gdb_interface.initialize_workflow(workflow)
        return workflow

    def execute_workflow(self):
        """Begin execution of a BEE workflow."""
        self._gdb_interface.execute_workflow()

    def pause_workflow(self):
        """Pause the execution of a BEE workflow."""
        self._gdb_interface.pause_workflow()

    def resume_workflow(self):
        """Resume the execution of a paused BEE workflow."""
        self._gdb_interface.resume_workflow()

    def reset_workflow(self, workflow_id):
        """Reset the execution state and ID of a BEE workflow."""
        self._workflow_id = workflow_id
        self._gdb_interface.reset_workflow(self._workflow_id)
        self._gdb_interface.set_workflow_state('SUBMITTED')

    def finalize_workflow(self):
        """Deconstruct a BEE workflow."""
        self._workflow_id = None
        self._gdb_interface.cleanup()

    def add_task(self, name, base_command, inputs, outputs, requirements=None, hints=None,
                 stdout=None, stderr=None):
        """Add a new task to a BEE workflow.

        :param name: the name given to the task
        :type name: str
        :param base_command: the base command for the task
        :type base_command: str or list of str
        :param requirements: the task-specific requirements
        :type requirements: list of Requirement
        :param hints: the task-specific hints (optional requirements)
        :type hints: list of Hint
        :param inputs: the task inputs
        :type inputs: list of StepInput
        :param outputs: the task outputs
        :type outputs: list of StepOutput
        :param stdout: the name of the file to which to redirect stdout
        :type stdout: str
        :param stderr: the name of the file to which to redirect stderr
        :type stderr: str
        :rtype: Task
        """
        # Immutable default arguments
        if inputs is None:
            inputs = []
        if outputs is None:
            outputs = []
        if requirements is None:
            requirements = []
        if hints is None:
            hints = []

        task = Task(name, base_command, hints, requirements, inputs, outputs, stdout, stderr,
                    self._workflow_id)
        # Load the new task into the graph database
        self._gdb_interface.load_task(task)
        return task

    def restart_task(self, task, checkpoint_file):
        """Restart a failed BEE workflow task.

        The task must have a beeflow:CheckpointRequirement hint. If there are no
        remaining retry attemps (num_tries = 0) then the graph database is
        unmodified and this method returns None.

        :param task: the task to restart
        :type task: Task
        :rtype: Task or None
        """
        for hint in task.hints:
            if hint.class_ == "beeflow:CheckpointRequirement":
                if hint.params["num_tries"] > 0:
                    hint.params["num_tries"] -= 1
                    hint.params["bee_checkpoint_file__"] = checkpoint_file
                    break
                state = self.get_task_state(task)
                self.set_task_state(task, f"FAILED RESTART: {state}")
                return None
        else:
            raise ValueError("invalid task for checkpoint restart")

        new_task = task.copy(new_id=True)
        # Pattern match on task name
        # Append (1) if not in name already
        # Increment number on each restart
        match = re.match(r".+\(([0-9]+)\)$", new_task.name)
        if not match:
            new_task.name += "(1)"
        else:
            i = int(match.group(1))
            new_task.name = re.sub(r"\([0-9]+\)", f"({i + 1})", new_task.name)
        metadata = self.get_task_metadata(task)
        self._gdb_interface.restart_task(task, new_task)
        self.set_task_metadata(new_task, metadata)
        self.set_task_state(task, "RESTARTED")
        self.set_task_state(new_task, "READY")
        return new_task

    def finalize_task(self, task):
        """Mark a BEE workflow task as completed.

        This method also automatically deduces what tasks are now
        runnable, updating their states to ready and returning a
        list of the runnable tasks.

        :param task: the task to finalize
        :type task: Task
        :rtype: list of Task
        """
        self._gdb_interface.finalize_task(task)
        self._gdb_interface.initialize_ready_tasks()
        return self._gdb_interface.get_ready_tasks()

    def get_task_by_id(self, task_id):
        """Get a task by its Task ID.

        :param task_id: the task's ID
        :type task_id: str
        :rtype: Task
        """
        return self._gdb_interface.get_task_by_id(task_id)

    def get_workflow(self):
        """Get a loaded BEE workflow.

        Returns a tuple of (workflow_description, tasks)

        :rtype: tuple of (Workflow, list of Task)
        """
        workflow = self._gdb_interface.get_workflow_description()
        tasks = self._gdb_interface.get_workflow_tasks()
        return workflow, tasks

    def get_workflow_outputs(self):
        """Get the outputs from a BEE workflow.

        :rtype: list of OutputParameter
        """
        workflow = self._gdb_interface.get_workflow_description()
        return workflow.outputs

    def get_workflow_state(self):
        """Get the value of the workflow state.

        :rtype: str
        """
        state = self._gdb_interface.get_workflow_state()
        return state

    def set_workflow_state(self, state):
        """Return workflow's current state.

        :param state: the new state of the workflow
        :type state: str
        """
        self._gdb_interface.set_workflow_state(state)

    def get_ready_tasks(self):
        """Get ready tasks from a BEE workflow.

        :rtype: list of Task
        """
        return self._gdb_interface.get_ready_tasks()

    def get_dependent_tasks(self, task):
        """Get the dependents of a task in a BEE workflow.

        :param task: the task whose dependents to retrieve
        :type task: Task
        :rtype: list of Task
        """
        return self._gdb_interface.get_dependent_tasks(task)

    def get_task_state(self, task):
        """Get the state of the task in a BEE workflow.

        :param task: the task whose state to retrieve
        :type task: Task
        :rtype: str
        """
        return self._gdb_interface.get_task_state(task)

    def set_task_state(self, task, state):
        """Set the state of the task in a BEE workflow.

        This method should not be used to set a task as completed.
        finalize_task() should instead be used.

        :param task: the task whose state to set
        :type task: Task
        :param state: the new state of the task
        :type state: str
        """
        self._gdb_interface.set_task_state(task, state)

    def get_task_metadata(self, task):
        """Get the job description metadata of a task in a BEE workflow.

        :param task: the task whose metadata to retrieve
        :type task: Task
        :rtype: dict
        """
        return self._gdb_interface.get_task_metadata(task)

    def set_task_metadata(self, task, metadata):
        """Set the job description metadata of a task in a BEE workflow.

        This method should not be used to update task state.
        set_task_state() or finalize_task() should instead be used.

        :param task: the task whose metadata to set
        :type task: Task
        :param metadata: the job description metadata
        :type metadata: dict
        """
        self._gdb_interface.set_task_metadata(task, metadata)

    def get_task_input(self, task, input_id):
        """Get a task input object.

        :param task: the task whose input to retrieve
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :rtype: StepInput
        """
        return self._gdb_interface.get_task_input(task, input_id)

    def set_task_input(self, task, input_id, value):
        """Set the value of a task input.

        :param task: the task whose input to set
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :param value: str or int or float
        """
        self._gdb_interface.set_task_input(task, input_id, value)

    def get_task_output(self, task, output_id):
        """Get a task output object.

        :param task: the task whose output to retrieve
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :rtype: StepOutput
        """
        return self._gdb_interface.get_task_output(task, output_id)

    def set_task_output(self, task, output_id, value):
        """Set the value of a task output.

        :param task: the task whose output to set
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :param value: the output value to set
        :type value: str or int or float
        """
        self._gdb_interface.set_task_output(task, output_id, value)

    def evaluate_expression(self, task, id_, output=False):
        """Evaluate a task input/output expression.

        Expression can be either a string concatenation in a StepInput
        valueFrom field or a parameter substitution in a StepOutput
        glob field. The only special variable supported in valueFrom is
        self.path.

        :param task: the task whose expression to evaluate
        :type task: Task
        :param id_: the id of the step input/output
        :type id_: str
        :param output: true if output glob expression being evaluated, else false
        :type output: bool
        """
        self._gdb_interface.evaluate_expression(task, id_, output)

    def workflow_completed(self):
        """Return true if all of a workflow's final tasks have completed, else false.

        :rtype: bool
        """
        return self._gdb_interface.workflow_completed()

    def workflow_initialized(self):
        """Return true if a workflow has been initialized, else false.

        Currently functionally the same as workflow_loaded() but may
        change when multiple workflows per database instance are supported.

        :rtype: bool
        """
        return self._gdb_interface.initialized()

    def workflow_loaded(self):
        """Return true if a workflow is loaded, else false.

        :rtype: bool
        """
        return bool(not self._gdb_interface.empty())

    @property
    def workflow_id(self):
        """Retrieve the workflow ID from the workflow interface.

        If workflow ID is not populated, this grabs it from the database.

        If no workflow is loaded, None is returned.
        :rtype: str
        """
        if self._workflow_id is None and self.workflow_loaded():
            workflow, _ = self.get_workflow()
            self._workflow_id = workflow.id

        return self._workflow_id



"""Cloud API tests."""
from beeflow.common.cloud import provider


def test_cloud_api():
    """Simple test to check the cloud provider API."""
    mock = provider.MockProvider()

    mock.setup_cloud('empty config....')

    assert mock.get_ext_ip_addr('some-node')



"""Unit tests for the workflow manager."""

import pathlib
import pytest
from beeflow.wf_manager.wf_manager import create_app
from beeflow.wf_manager.resources import wf_utils
from beeflow.wf_manager.common import wf_db
from beeflow.tests.mocks import MockWFI, MockCwlParser, MockGDBInterface
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.wf_interface import WorkflowInterface

# We use this as the test workflow id
WF_ID = '42'


@pytest.fixture
def app():
    """Create a new flask app object."""
    flask_app = create_app()
    flask_app.config.update({
        "TESTING": True
    })

    yield flask_app


@pytest.fixture()
def client(app):
    """Return the flask test client."""
    return app.test_client


@pytest.fixture()
def teardown_workflow():
    """Just tear down for tests that use the workflow directory."""
    yield
    wf_utils.remove_wf_dir(WF_ID)


@pytest.fixture()
def setup_teardown_workflow(teardown_workflow):
    """Set up and tear down for tests that use the workflow directory."""
    wf_utils.create_workflow_dir(WF_ID)
    wf_utils.create_wf_status(WF_ID)
    yield


def _url():
    """Return URL to the workflow manager end point."""
    wfm_listen_port = bc.get('workflow_manager', 'listen_port')
    workflow_manager = 'bee_wfm/v1/jobs'
    return f'http://127.0.0.1:{wfm_listen_port}/{workflow_manager}/'


def _resource(tag=""):
    """Return the url for the workflow manager endpoint."""
    return _url() + str(tag)


# WFList Tests
def test_submit_workflow(client, mocker, teardown_workflow):
    """Test submitting a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_list.CwlParser', new=MockCwlParser)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.create_image',
                 return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.start_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.wait_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.kill_gdb', return_value=True)
    mocker.patch('beeflow.common.wf_data.generate_workflow_id', return_value='42')
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=WorkflowInterface(MockGDBInterface()))
    mocker.patch('subprocess.run', return_value=True)
    script_path = pathlib.Path(__file__).parent.resolve()
    tarball = script_path / 'clamr-wf.tgz'
    with open(tarball, 'rb') as tarball_contents:
        resp = client().post('/bee_wfm/v1/jobs/', data={
            'wf_name': 'clamr',
            'wf_filename': tarball,
            'main_cwl': 'clamr_wf.cwl',
            'yaml': 'clamr_job.yml',
            'workflow_archive': tarball_contents,
            'workdir': '.'
        })

    # Remove task added during the test
    wf_db.delete_task(42, WF_ID)
    print(resp.json)
    assert resp.json['msg'] == 'Workflow uploaded'


def test_reexecute_workflow(client, mocker, teardown_workflow):
    """Test reexecuting a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_list.CwlParser', new=MockCwlParser)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.create_image',
                 return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.start_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.wait_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_list.dep_manager.kill_gdb', return_value=True)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.common.wf_data.generate_workflow_id', return_value='42')
    mocker.patch('subprocess.run', return_value=True)

    script_path = pathlib.Path(__file__).parent.resolve()
    tarball = script_path / '42.tgz'
    with open(tarball, 'rb') as tarball_contents:
        resp = client().put('/bee_wfm/v1/jobs/', data={
                            'wf_filename': tarball,
                            'wf_name': 'clamr',
                            'workflow_archive': tarball_contents,
                            'workdir': '.'
                            })

        print(resp.json)
        assert resp.json['msg'] == 'Workflow uploaded'


# WFActions Tests
def test_start_workflow(client, mocker):
    """Test starting a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_tm', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_scheduler', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.update_wf_status', return_value=None)
    resp = client().post(f'/bee_wfm/v1/jobs/{WF_ID}')
    assert resp.status_code == 200


def test_workflow_status(client, mocker, setup_teardown_workflow):
    """Test getting workflow status."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    wf_name = 'wf'
    wf_status = 'Pending'
    bolt_port = 3030
    gdb_pid = 12345
    wf_db.add_workflow(WF_ID, wf_name, wf_status, 'dir', bolt_port, gdb_pid)
    wf_db.add_task(123, WF_ID, 'task', "WAITING")
    wf_db.add_task(124, WF_ID, 'task', "RUNNING")
    resp = client().get(f'/bee_wfm/v1/jobs/{WF_ID}')
    wf_db.delete_workflow(WF_ID)
    wf_db.delete_task(123, WF_ID)
    wf_db.delete_task(124, WF_ID)
    assert 'RUNNING' in resp.json['tasks_status']


def test_cancel_workflow(client, mocker, setup_teardown_workflow):
    """Test cancelling a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())

    request = {'wf_id': WF_ID}
    resp = client().delete(f'/bee_wfm/v1/jobs/{WF_ID}', json=request)
    assert resp.json['status'] == 'Cancelled'
    assert resp.status_code == 202


def test_pause_workflow(client, mocker, setup_teardown_workflow):
    """Test pausing a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.tests.mocks.MockWFI.get_workflow_state',
                 return_value='RUNNING')

    wf_utils.update_wf_status(WF_ID, 'Running')
    request = {'option': 'pause'}
    resp = client().patch(f'/bee_wfm/v1/jobs/{WF_ID}', json=request)
    assert resp.json['status'] == 'Workflow Paused'
    assert resp.status_code == 200


def test_resume_workflow(client, mocker, setup_teardown_workflow):
    """Test resuming a workflow."""
    mocker.patch('beeflow.wf_manager.resources.wf_utils.get_workflow_interface',
                 return_value=MockWFI())
    mocker.patch('beeflow.tests.mocks.MockWFI.get_workflow_state',
                 return_value='PAUSED')
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_tm', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.submit_tasks_scheduler', return_value=None)
    mocker.patch('beeflow.wf_manager.resources.wf_utils.update_wf_status', return_value=None)

    wf_utils.update_wf_status(WF_ID, 'Paused')
    request = {'option': 'resume'}
    resp = client().patch(f'/bee_wfm/v1/jobs/{WF_ID}', json=request)
    assert resp.json['status'] == 'Workflow Resumed'
    assert resp.status_code == 200
# pylama:ignore=W0613,W0621



"""Serialization module.

This holds a class for simple serialization of Python objects.
"""
import abc


class Serializable(abc.ABC):
    """Serializable base class.

    This class allows subclasses to easily serialize into simple Python
    data types which can be serialized into JSON.
    """

    def encode(self):
        """Encode and return a simple Python data type.

        Produce a simple Python data type for serialization.
        """
        return self.__dict__

    @staticmethod
    @abc.abstractmethod
    def decode(data):
        """Decode a serialized object and return an instance.

        Decode a simple Python data type and return and instance of
        the object.
        """



"""BEE Cloud Installer Script."""
import argparse
import subprocess
import sys
import time
import yaml
import jinja2
import requests

from beeflow.common import cloud
from beeflow.common.config_driver import BeeConfig as bc


def run(private_key_file, bee_user, ip_addr, cmd):
    """Run a command on the remote host."""
    proc = subprocess.run([
        'ssh',
        '-i', private_key_file,
        '-o', 'StrictHostKeyChecking=no',
        '-o', 'ConnectTimeout=8',
        f'{bee_user}@{ip_addr}',
        cmd,
    ], check=True)
    return proc.returncode


def scp(bee_user, ip_addr, priv_key_file, src, dst):
    """SCP a file in src to dst on the remote machine."""
    proc = subprocess.run([
        'scp',
        '-i', priv_key_file,
        '-o', 'StrictHostKeyChecking=no',
        src,
        f'{bee_user}@{ip_addr}:{dst}',
    ], check=True)
    if proc.returncode != 0:
        raise RuntimeError(f'Could not copy file "{src}" to "{dst}" on the remote machine')


def copy_files_to_instance(provider, bee_user, private_key_file, head_node,
                           copy_files):
    """Copy files over to the instance."""
    print('Starting file copy step')
    ip_addr = provider.get_ext_ip_addr(head_node)
    for file in copy_files:
        src = file['src']
        dst = file['dst']
        scp(bee_user, ip_addr, private_key_file, src, dst)
    print('Finished')


def launch_tm(provider, private_key_file, bee_user, launch_cmd, head_node,
              tm_listen_port, wfm_listen_port):
    """Start the Task Manager on the remote head node."""
    print('Launching the Remote Task Manager')
    ip_addr = provider.get_ext_ip_addr(head_node)
    tm_proc = None

    try:
        # Now start the Task Manager
        tm_proc = subprocess.Popen([
            'ssh',
            '-i', private_key_file,
            '-o', 'StrictHostKeyChecking=no',
            '-o', 'ConnectTimeout=8',
            f'{bee_user}@{ip_addr}',
            launch_cmd,
        ])
        time.sleep(10)
        if tm_proc.poll() is not None:
            raise RuntimeError('Failed to launch the Remote Task Manager')

        # Set up the connection with an SSH tunnel
        subprocess.run([
            'ssh',
            f'{bee_user}@{ip_addr}',
            '-i', private_key_file,
            # The TM is listening on the remote machine
            '-L', f'{tm_listen_port}:localhost:{tm_listen_port}',
            # The WFM is listening on this machine
            '-R', f'{wfm_listen_port}:localhost:{wfm_listen_port}',
            '-N',
            '-o', 'ExitOnForwardFailure=yes',
            '-o', 'StrictHostKeyChecking=no',
            # Required in order to allow port forwarding
            '-o', 'UserKnownHostsFile=/dev/null',
        ], check=True)
    except KeyboardInterrupt:
        print('Got keyboard interrupt, quitting')
    finally:
        if tm_proc is not None:
            print('Killing the task manager')
            # TODO: This could be done much better using a pidfile
            run(private_key_file, bee_user, ip_addr, 'pkill python')


def connect(provider, private_key_file, bee_user, head_node,
            tm_listen_port, wfm_listen_port, max_retries):
    """Connect to an already running TM."""
    ip_addr = provider.get_ext_ip_addr(head_node)
    tun_proc = subprocess.Popen([
        'ssh',
        f'{bee_user}@{ip_addr}',
        '-i', private_key_file,
        # The TM is listening on the remote machine
        '-L', f'{tm_listen_port}:localhost:{tm_listen_port}',
        # The WFM is listening on this machine
        '-R', f'{wfm_listen_port}:localhost:{wfm_listen_port}',
        '-N',
        '-o', 'ExitOnForwardFailure=yes',
        '-o', 'StrictHostKeyChecking=no',
        # Required in order to allow port forwarding
        '-o', 'UserKnownHostsFile=/dev/null',
    ])
    time.sleep(3)
    returncode = tun_proc.poll()
    # Ensure the SSH process is still running
    if returncode is not None:
        sys.exit(f'Tunnel set up failed with error: {returncode}')
    # Now continue until we can ping the Task Manager
    url = f'http://localhost:{tm_listen_port}/status'
    interval = 5
    # Keep trying to get a status from the Task Manager (this could probably be
    # done better with some sort of backoff algorithm)
    for _ in range(max_retries):
        print('Trying to connect to the TM')
        time.sleep(interval)
        try:
            resp = requests.get(url, timeout=60)
            status = resp.json()
            print(f'Connected to the TM with status "{status}"')
            return
        except requests.ConnectionError:
            pass
    # If we get here, then set up might be taking longer than interval *
    # max_retries or the set up failed somehow
    tun_proc.kill()
    sys.exit('Connection attempt to the TM failed')


def main():
    """Cloud launcher entry point."""
    # Argument parsing
    parser = argparse.ArgumentParser(description='BEE Cloud Installer')
    parser.add_argument('provider_config', help='provider config yaml file')
    parser.add_argument('--config_file', help='bee.conf file')
    parser.add_argument('--setup-cloud', action='store_true', help='set up the remote cloud')
    parser.add_argument('--copy', action='store_true', help='copy over files in the config')
    parser.add_argument('--tm', action='store_true', help='start the TM')
    parser.add_argument('--connect', action='store_true', help='connect to an already running TM')
    parser.add_argument('--debug', action='store_true',
                        help='debug the cloud template, don\'t make any API calls')
    parser.add_argument('--max-retries', default=124, type=int,
                        help='max number of retry attempts when used with the `--connect` option')
    args = parser.parse_args()

    # Get configuration information
    if args.config_file is not None:
        bc.init(userconfig=args.config_file, workload_scheduler='Simple')
    else:
        bc.init(workload_scheduler='Simple')
    # Load the provider config file
    with open(args.provider_config, 'r', encoding='utf-8') as fp:
        cfg = yaml.load(fp, Loader=yaml.Loader)

    # Get the component ports for forwarding connections
    wfm_listen_port = cfg['wfm_listen_port']
    tm_listen_port = cfg['tm_listen_port']
    private_key_file = cfg['private_key_file']
    bee_user = cfg['bee_user']
    launch_cmd = cfg['tm_launch_cmd']
    head_node = cfg['head_node']
    template_file = cfg['template_file']
    copy_files = cfg['copy_files']
    provider = cfg['provider']
    provider = cloud.get_provider(provider, **cfg)

    if args.setup_cloud:
        print('Creating cloud from template...')
        with open(template_file, encoding='utf-8') as fp:
            tmpl = jinja2.Template(fp.read())
        tmpl_data = tmpl.render(**cfg)
        if args.debug:
            # Display the template
            sys.stdout.write(tmpl_data)
            sys.exit()
        provider.setup_cloud(tmpl_data)
        time.sleep(20)
        print('Setup complete')
    if args.copy and copy_files is not None:
        copy_files_to_instance(provider=provider, bee_user=bee_user,
                               private_key_file=private_key_file,
                               head_node=head_node, copy_files=copy_files)
    if args.tm:
        launch_tm(provider=provider, private_key_file=private_key_file,
                  bee_user=bee_user, launch_cmd=launch_cmd, head_node=head_node,
                  wfm_listen_port=wfm_listen_port,
                  tm_listen_port=tm_listen_port)
    if args.connect:
        connect(provider=provider, private_key_file=private_key_file,
                bee_user=bee_user, head_node=head_node,
                wfm_listen_port=wfm_listen_port, tm_listen_port=tm_listen_port,
                max_retries=args.max_retries)


if __name__ == '__main__':
    main()
# Ignore R1732: Using a 'with' statement for resource allocation doesn't really apply here. We'd
#               need to completely restructure the code to use it, but it's something to think
#               about later on.
# Ignore W0511: This TODO needs to be addressed later on, perhaps with a restructure of this code.
# pylama:ignore=R1732,W0511



"""Helper functions for starting and stopping the GDB."""
import time
from beeflow.wf_manager.common import dep_manager
from beeflow.wf_manager.resources import wf_utils


def start():
    """Start the GDB."""
    # dep_manager.kill_gdb()
    # dep_manager.remove_current_run()
    try:
        dep_manager.create_image()
    except dep_manager.NoContainerRuntime:
        raise RuntimeError('Charliecloud is not installed') from None
    bolt_port = wf_utils.get_open_port()
    http_port = wf_utils.get_open_port()
    https_port = wf_utils.get_open_port()
    pid = dep_manager.start_gdb('/', bolt_port, http_port, https_port)
    time.sleep(10)
    return pid


def stop(pid):
    """Stop the GDB."""
    dep_manager.kill_gdb(pid)
    # dep_manager.remove_current_run()



"""Mocks for the WFM and TM tests."""

from copy import deepcopy
from beeflow.common.wf_data import StepInput, StepOutput
from beeflow.common import expr


class MockTask:
    """Mock task object."""

    def __init__(self, task_name="task", state="DEFAULT", new_id=42):
        """Initialize."""
        self.name = task_name
        self.state = state
        self.id = new_id
        self.metadata = {}


class MockWFI:
    """Mock worfklow interface object."""

    def __init__(self):
        """Set the fake workflow id."""
        self._workflow_id = '42'
        self._loaded = False

    def pause_workflow(self):
        """Pause a workflow."""
        return

    def resume_workflow(self):
        """Resume a workflow."""
        return

    def reset_workflow(self, wf_id): #noqa
        """Reset a workflow."""
        wf_id = 0 # noqa

    def get_dependent_tasks(self, task): # noqa
        """Get depdendent states."""
        return [MockTask()]

    def get_task_metadata(self, task):
        """Get the task's metadata."""
        return task.metadata

    def set_task_metadata(self, task, metadata):
        """Set the metadata for this task."""
        task.metadata = metadata

    def get_task_by_id(self, task_id): # noqa
        """Return a mock task from an ID."""
        return MockTask()

    def get_workflow_state(self):
        """Get a workflow state."""
        return 'PENDING'

    @property
    def workflow_id(self):
        """Return the workflow id."""
        return self._workflow_id

    def get_ready_tasks(self):
        """Return a mock ready task."""
        return [MockTask()]

    def workflow_initialized(self):
        """Fake that the workflow has been initialized."""

    def set_task_state(self, task, job_state): # noqa
        """Set the state of a task."""
        task.state = job_state

    def workflow_loaded(self):
        """Fake workflow being loaded."""
        return self._loaded

    def initialize_workflow(self, wf_id, wf_name, inputs, outputs, req=None, hints=None): # noqa
        """Initialize the workflow."""
        self._loaded = True

    def finalize_workflow(self):
        """Finalize the workflow."""
        self._loaded = False

    def create_requirement(self, req_class, key, value):
        """Fake creating a requirement."""

    def add_task(self, name, base_command, inputs, outputs, requirements, hints, stdout, stderr):
        """Fake adding a task."""

    def get_workflow(self):
        """Get a list of workflows."""
        return None, [MockTask("task1"), MockTask("task2")]

    def get_task_state(self, task_name): # noqa
        """Returns the task state."""
        return "RUNNING"

    def execute_workflow(self):
        """Fake executing a workflow."""
        pass # noqa 


class MockGDBInterface:
    """A mock GDB interface.

    TODO: This is very crude, and will likely break with more complicated GDB
    interactions.
    """

    def __init__(self, **_kwargs):
        """Create a new mock gdb (ignore kwargs)."""
        self.workflow = None
        self.workflow_state = None
        self.tasks = {}
        self.task_states = {}
        self.task_metadata = {}
        self.inputs = {}
        self.outputs = {}

    def connect(self, **kwargs):
        """Initialize a graph database interface with a driver."""

    def initialize_workflow(self, workflow):
        """Begin construction of a workflow in the graph database."""
        self.workflow = workflow
        self.workflow_state = 'SUBMITTED'

    def _is_ready(self, task_id):
        """Check whether a given task is ready to run."""
        sources = [inp.source for inp in self.inputs[task_id].values()]
        task_deps = [task_id for task_id in self.tasks
                     if any(source in self.outputs[task_id] for source in sources)]
        return all(self.task_states[task_dep_id] == 'COMPLETED'
                   for task_dep_id in task_deps)

    def execute_workflow(self):
        """Begin execution of the loaded workflow."""
        self.workflow_state = 'RUNNING'
        for task_id in self.task_states:
            if self._is_ready(task_id):
                self.task_states[task_id] = 'READY'

    def pause_workflow(self):
        """Pause execution of a running workflow."""
        self.workflow_state = 'PAUSED'

    def resume_workflow(self):
        """Resume execution of a running workflow."""
        self.workflow_state = 'RESUME'

    def reset_workflow(self, new_id):
        """Reset the execution state and ID of a workflow."""
        self.workflow = deepcopy(self.workflow)
        self.workflow.id = new_id
        for task_id, task in self.tasks.items():
            task.workflow_id = new_id
            self.task_metadata[task_id] = {}
            self.task_states[task_id] = 'WAITING'

    def load_task(self, task):
        """Load a task into a workflow in the graph database."""
        self.tasks[task.id] = task
        self.task_states[task.id] = 'WAITING'
        self.task_metadata[task.id] = {}
        self.inputs[task.id] = {}
        self.outputs[task.id] = {}
        for inp in task.inputs:
            self.inputs[task.id][inp.id] = inp
        for outp in task.outputs:
            self.outputs[task.id][outp.id] = outp

    def initialize_ready_tasks(self):
        """Set runnable tasks in a workflow to ready."""
        for task_id in self.tasks:
            if self._is_ready(task_id) and self.task_states[task_id] == 'WAITING':
                self.task_states[task_id] = 'READY'

    def restart_task(self, _old_task, new_task):
        """Create a new task from a failed task checkpoint restart enabled."""
        self.load_task(new_task)

    def finalize_task(self, task):
        """Set a task's state to completed."""
        self.task_states[task.id] = 'COMPLETED'

    def get_task_by_id(self, task_id):
        """Return a workflow Task given its ID."""
        return self.tasks[task_id]

    def get_workflow_description(self):
        """Return the workflow description from the graph database."""
        return deepcopy(self.workflow)

    def get_workflow_state(self):
        """Return workflow's current state."""
        return self.workflow_state

    def set_workflow_state(self, state):
        """Return workflow's current state."""
        self.workflow_state = state

    def get_workflow_tasks(self):
        """Return a workflow's tasks from the graph database."""
        return list(self.tasks.values())

    def get_workflow_requirements_and_hints(self):
        """Return a tuple containing a list of requirements and a list of hints."""
        return (None, None)

    def get_ready_tasks(self):
        """Return the tasks in a workflow with state 'READY'."""
        return [task for task_id, task in self.tasks.items()
                if self.task_states[task_id] == 'READY']

    def get_dependent_tasks(self, task):
        """Return the dependents of a task in a workflow."""
        tasks = [dep_task for dep_task_id, dep_task in self.tasks.items()
                 if any(outp_id in [inp.source
                                    for inp in self.inputs[dep_task_id].values()]
                        for outp_id in self.outputs[task.id])]
        return tasks

    def get_task_state(self, task):
        """Return the state of a task."""
        return self.task_states[task.id]

    def set_task_state(self, task, state):
        """Set the state of a task."""
        self.task_states[task.id] = state

    def get_task_metadata(self, task):
        """Return the job description metadata of a task."""
        return self.task_metadata[task.id]

    def set_task_metadata(self, task, metadata):
        """Set the job description metadata of a task."""
        self.task_metadata[task.id] = metadata

    def get_task_input(self, task, input_id):
        """Get a task input object."""
        inp = self.inputs[task.id][input_id]
        try:
            inp.id # noqa (trying to get an AttributeError here)
            return inp
        except AttributeError:
            return StepInput(input_id, 'File', inp,
                             'default.txt', input_id, None, None, None)

    def set_task_input(self, task, input_id, value):
        """Set the value of a task input."""
        self.inputs[task.id][input_id] = value

    def get_task_output(self, task, output_id):
        """Get a task output object."""
        return self.outputs[task.id][output_id]

    def set_task_output(self, task, output_id, value):
        """Set the value of a task output."""
        self.outputs[task.id][output_id] = StepOutput(
            output_id, 'File', value, value,
        )

    def evaluate_expression(self, task, id_, output):
        """Evaluate a task input/output expression."""
        input_pairs = {id_: inp.value for id_, inp in self.inputs[task.id].items()}
        if output:
            step_outp = self.outputs[task.id][id_]
            val = expr.eval_output(input_pairs, step_outp.glob)
            if val is not None:
                self.outputs[task.id][id_] = StepOutput(step_outp.id, step_outp.type,
                                                        step_outp.value, val)
        else:
            step_inp = self.inputs[task.id][id_]
            val = expr.eval_input(input_pairs, step_inp.value_from)
            self.inputs[task.id][id_] = StepInput(step_inp.id, 'string', val,
                                                  step_inp.default,
                                                  step_inp.source,
                                                  step_inp.prefix,
                                                  step_inp.position,
                                                  step_inp.value_from)

    def workflow_completed(self):
        """Return true if all of a workflow's final tasks have completed, else false."""
        return all(state == 'COMPLETED' for state in self.task_states.values())

    def initialized(self):
        """Return true if the database connection has been initialized, else false."""
        return not self.empty()

    def empty(self):
        """Return true if the graph database is empty, else false."""
        return self.workflow is None

    def cleanup(self):
        """Clean up all data in the graph database."""
        self.workflow = None
        self.workflow_state = None
        self.tasks.clear()
        self.task_states.clear()
        self.task_metadata.clear()
        self.inputs.clear()
        self.outputs.clear()

    def close(self):
        """Close the connection to the graph database."""


def mock_create_image(): # noqa
    """Fake image creation."""
    pass # noqa


class MockCwlParser:
    """Mock the CWLParser."""

    def __init__(self, bolt_port):
        """Need a port."""
        self.bolt_port = bolt_port

    def parse_workflow(self, wf_id, cwl_path, yaml_file=None): # noqa
        """Parse the workflow."""
        return MockWFI()


class MockWorkerSubmission:
    """Mock Worker during submission."""

    def submit_task(self, task): # noqa
        """Return submission."""
        return 1, 'PENDING'

    def query_task(self, job_id): #noqa
        """Return state of task."""
        return 'RUNNING'

    def cancel_task(self, job_id): # noqa
        """Return cancelled status"""
        return 'CANCELLED'


class MockWorkerCompletion:
    """Mock Worker after completion."""

    def submit_task(self, task): #noqa
        """Submit a task."""
        return 1, 'PENDING'

    def query_task(self, job_id): #noqa
        """Submit a task."""
        return 'COMPLETED'

    def cancel_task(self, job_id): #noqa
        """Cancel a task."""
        return 'CANCELLED'


class MockResponse:
    """Mock a response."""

    def __init__(self, status_code):
        """Initialize response."""
        self.status_code = status_code

    @staticmethod
    def json():
        """Return json output."""
        return '{}'


def mock_put(url, params=None, **kwargs): # noqa
    """Fake put."""
    return MockResponse(200)


def mock_post(url, params=None, **kwargs): # noqa
    """Fake post."""
    return MockResponse(200)


def mock_get(url, params=None, **kwargs): # noqa
    """Fake get."""
    return MockResponse(200)


def mock_delete(url, params=None, **kwargs): # noqa
    """Fake delete."""
    return MockResponse(200)



"""Provider class, representing a specific Cloud provider."""
import abc


class Provider(abc.ABC):
    """Provider Abstract Base Class."""

    @abc.abstractmethod
    def get_ext_ip_addr(self, node_name):
        """Get the external IP address of the node, if it has one."""

    @abc.abstractmethod
    def setup_cloud(self, config):
        """Set up the cloud based on the config data."""


class MockProvider(Provider):
    """Mock provider class for testing."""

    def __init__(self, **kwargs):
        """Construct a mock provider."""

    def get_ext_ip_addr(self, node_name):
        """Get the external IP address of the node, if it has one."""
        return '100.100.100.100'

    def setup_cloud(self, config):
        """Set up the cloud based on the config data."""



"""Scheduler types for storing scheduling information.

This file holds classes for respresenting workflows and tasks/jobs, as
well as resources that will be used during scheduling.
"""
from beeflow.scheduler import serializable
from beeflow.scheduler import resource_allocation


class Task(serializable.Serializable):
    """Representation of a Task and its various requirements.

    This class represents the task as a set of resource requirements
    which can be used by the scheduling algorithm to easily determine
    the best allocation.
    """

    def __init__(self, workflow_name, task_name, requirements=None,
                 allocations=None):
        """Task constructor.

        Create a new Task with given parameters.
        :param workflow_name: name of the workflow
        :type workflow_name: str
        :param task_name: name of the task
        :type task_name: str
        :param requirements: requirements dict
        :type requirements: dict of requirements
        :param allocations: list of current allocations
        :type allocations: list of instance of Allocation
        """
        self.workflow_name = workflow_name
        self.task_name = task_name
        self.requirements = (
            resource_allocation.Requirements.decode(requirements)
            if requirements is not None else {}
        )
        self.allocations = allocations if allocations is not None else []

    def encode(self):
        """Encode and return a simple Python data type.

        Produce a simple Python data type for serialization.
        """
        result = dict(self.__dict__)
        if 'allocations' in result and result['allocations']:
            result['allocations'] = [alloc.encode()
                                     for alloc in result['allocations']]
        if 'requirements' in result and result['requirements'] is not None:
            # result['requirements'] = result['requirements'].__dict__
            result['requirements'] = result['requirements'].encode()
        return result

    @staticmethod
    def decode(data):
        """Decode a serialized object and return an instance.

        Decode a simple Python data type and return and instance of
        the object.
        """
        data = dict(data)
        return Task(**data)



"""Container build driver.

All container-based build systems belong here.
"""

import os
import shutil
import subprocess
import tempfile
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common import log as bee_logging
from beeflow.common.build.build_driver import BuildDriver
from beeflow.common.crt.charliecloud_driver import CharliecloudDriver as crt_driver


log = bee_logging.setup(__name__)


class ContainerBuildDriver(BuildDriver):
    """Driver interface between WFM and a container build system.

    A driver object must implement an __init__ method that
    requests a RTE, and a method to return the requested RTE.
    """


class CharliecloudBuildDriver(ContainerBuildDriver):
    """Driver interface between WFM and a container build system.

    A driver object must implement an __init__ method that
    requests a RTE, and a method to return the requested RTE.
    """

    def __init__(self, task):
        """Begin build request.

        Parse hints and requirements to determine target build
        system. Harvest relevant BeeConfig params in kwargs.

        :param requirements: the workflow requirements
        :type requirements: set of Requirement instances
        :param hints: the workflow hints (optional requirements)
        :type hints: set of Requirement instances
        :param kwargs: Dictionary of build system config options
        :type kwargs: set of build system parameters
        """
        # Store build container archive pased on config file or relative to bee_workdir if not set.
        container_archive = bc.get('builder', 'container_archive')
        self.container_archive = bc.resolve_path(container_archive)
        os.makedirs(self.container_archive, exist_ok=True)
        # Deploy build tarballs relative to /var/tmp/username/beeflow by default
        deployed_image_root = bc.get('builder', 'deployed_image_root')
        # Make sure conf_file path exists
        os.makedirs(deployed_image_root, exist_ok=True)
        # Make sure path is absolute
        deployed_image_root = bc.resolve_path(deployed_image_root)
        self.deployed_image_root = deployed_image_root
        os.makedirs(self.deployed_image_root, exist_ok=True)
        # Set container-relative output directory based on BeeConfig, or use '/'
        container_output_path = bc.get('builder', 'container_output_path')
        self.container_output_path = container_output_path
        # record that a Charliecloud builder was used
        # bc.modify_section('user', 'builder', {'container_type': 'charliecloud'})
        self.task = task
        self.docker_image_id = None
        self.container_name = None
        docker_requirements = set()
        try:
            requirement_docker_requirements = self.task.requirements['DockerRequirement'].keys()
            docker_requirements = docker_requirements.union(requirement_docker_requirements)
            req_string = (f'{set(requirement_docker_requirements)}')
            log.info(f'task {self.task.id} requirement DockerRequirements: {req_string}')
        except (TypeError, KeyError):
            log.info(f'task {self.task.name} {self.task.id} no DockerRequirements in requirement')
        try:
            hint_docker_requirements = self.task.hints['DockerRequirement'].keys()
            docker_requirements = docker_requirements.union(hint_docker_requirements)
            hint_str = (f'{set(hint_docker_requirements)}')
            log.info(f'task {self.task.name} {self.task.id} hint DockerRequirements: {hint_str}')
        except (TypeError, KeyError):
            log.info(f'task {self.task.name} {self.task.id} hints has no DockerRequirements')
        log.info(f'task {self.task.id} union DockerRequirements : {docker_requirements}')
        exec_superset = self.resolve_priority()
        self.exec_list = [i for i in exec_superset if i[1] in docker_requirements]
        log_exec_list = [i[1] for i in self.exec_list]
        log.info(f'task {self.task.id} DockerRequirement execution order will be: {log_exec_list}')
        log.info('Execution order pre-empts hint/requirement status.')

    def get_docker_req(self, docker_req_param):
        """Get dockerRequirement, prioritizing requirements over hints.

        :param docker_req_param: the dockerRequirement parameter (e.g. 'dockerFile')
        :type docker_req_param: str

        When requirements are specified hints will be ignored.
        By default, tasks need not specify hints or requirements
        """
        task_docker_req = None
        # Get value if specified in requirements
        try:
            # Try to get Requirements
            task_docker_req = self.task.requirements['DockerRequirement'][docker_req_param]
        except (KeyError, TypeError):
            # Task Requirements are not mandatory. No docker_req_param specified in task reqs.
            task_docker_req = None
        # Ignore hints if requirements available
        if not task_docker_req:
            # Get value if specified in hints
            try:
                # Try to get Hints
                task_docker_req = self.task.hints['DockerRequirement'][docker_req_param]
            except (KeyError, TypeError):
                # Task Hints are not mandatory. No docker_req_param specified in task hints.
                task_docker_req = None
        return task_docker_req

    def process_docker_pull(self, addr=None, force=False):
        """Get the CWL compliant dockerPull dockerRequirement.

        CWL spec 09-23-2020: Specify a Docker image to
        retrieve using docker pull. Can contain the immutable
        digest to ensure an exact container is used.
        """
        task_addr = self.get_docker_req('dockerPull')

        # Use task specified image if image parameter empty
        if not addr:
            addr = task_addr

        # If Requirement is set but not specified, and param empty, do nothing and error.
        if self.task.requirements == {} and not addr:
            log.error("dockerPull set but no image path specified.")
            return 1
        # If no image specified and no image required, nothing to do.
        if not task_addr and not addr:
            log.info('No image specified and no image required, nothing to do.')
            return 0

        # Determine name for successful build target
        ch_build_addr = addr.replace('/', '%')

        ch_build_target = '/'.join([self.container_archive, ch_build_addr]) + '.tar.gz'
        # Return if image already exist and force==False.
        if os.path.exists(ch_build_target) and not force:
            log.info('Image already exists. If you want to refresh container, use force option.')
            log.info(f'Image path: {ch_build_target}')
            return 0
        # Force remove any cached images if force==True
        if os.path.exists(ch_build_target) and force:
            try:
                os.remove(ch_build_target)
            except FileNotFoundError:
                pass
            try:
                shutil.rmtree('/var/tmp/' + os.getlogin() + '/ch-image/' + ch_build_addr)
            except FileNotFoundError:
                pass

        # Out of excuses. Pull the image.
        cmd = (f'ch-image pull {addr}\n'
               f'ch-convert -i ch-image -o tar {ch_build_addr}'
               f' {self.container_archive}/{ch_build_addr}.tar.gz'
               )
        return subprocess.run(cmd, check=True, shell=True)

    def process_docker_load(self):
        """Get and process the CWL compliant dockerLoad dockerRequirment.

        CWL spec 09-23-2020: Specify a HTTP URL from which to
        download a Docker image using docker load.
        """
        # Need to know if dockerLoad is specified in order to determine fail/success
        req_dockerload = self.get_docker_req('dockerLoad')
        log.warning('Charliecloud does not have the concept of a layered image tarball.')
        log.warning('Did you mean to use dockerImport?')
        if req_dockerload:
            log.warning('dockerLoad specified as requirement.')
            return 1
        return 0

    def process_docker_file(self, task_dockerfile=None, force=False):
        """Get and process the CWL compliant dockerFile dockerRequirement.

        CWL spec 09-23-2020: Supply the contents of a Dockerfile
        which will be built using docker build. We have discussed implementing CWL
        change to expect a file handle instead of file contents, and use the file
        handle expectation here.
        """
        # beeflow:containerName is always processed before dockerFile, so safe to assume it exists
        # otherwise, raise an error.
        if self.container_name is None:
            log.error("dockerFile may not be specified without beeflow:containerName")
            return 1

        # Need dockerfile in order to build, else fail
        task_dockerfile = self.get_docker_req('dockerFile')
        if not task_dockerfile:
            log.error("dockerFile not specified as task attribute or parameter.")
            return 1

        # Create context directory to use as Dockerfile context, use container name so user
        # can prep the directory with COPY sources as needed.
        context_dir = os.path.expanduser('~')
        log.info(f'Context directory will be {context_dir}')

        # Determine name for successful build target
        ch_build_addr = self.container_name.replace('/', '%')

        ch_build_target = '/'.join([self.container_archive, ch_build_addr]) + '.tar.gz'
        log.info(f'Build will create tar ball at {ch_build_target}')
        # Return if image already exist and force==False.
        if os.path.exists(ch_build_target) and not force:
            return 0
        # Force remove any cached images if force==True
        if os.path.exists(ch_build_target) and force:
            try:
                os.remove(ch_build_target)
            except FileNotFoundError:
                pass
            try:
                shutil.rmtree('/var/tmp/' + os.getlogin() + '/ch-image/' + ch_build_addr)
            except FileNotFoundError:
                pass

        # Out of excuses. Build the image.
        with tempfile.NamedTemporaryFile(mode='w+', encoding='utf-8') as tmp:
            # Write the dockerfile to the tempfile
            tmp.write(task_dockerfile)
            tmp.flush()
            dockerfile_path = tmp.name
            # Build and run the command
            log.info('Context directory configured. Beginning build.')
            cmd = (f'ch-image build -t {self.container_name} --force '
                   f'-f {dockerfile_path} {context_dir}\n'
                   f'ch-convert -i ch-image -o tar {ch_build_addr} '
                   f'{self.container_archive}/{ch_build_addr}.tar.gz'
                   )
            log.info(f'Executing: {cmd}')
            return subprocess.run(cmd, check=True, shell=True)

    def process_docker_import(self, param_import=None):
        """Get and process the CWL compliant dockerImport dockerRequirement.

        CWL spec 09-23-2020: Provide HTTP URL to download and
        gunzip a Docker images using docker import. The param_import
        may be used to override DockerReuirement specs.
        """
        # If parameter import is specified, use that, otherwise look at task.
        if param_import:
            import_input_path = param_import
        else:
            # Get path for tarball to import
            import_input_path = self.get_docker_req('dockerImport')

        # Pull the image.
        file_name = crt_driver.get_ccname(import_input_path)
        cmd = (f'ch-convert {import_input_path} {self.deployed_image_root}/{file_name}')
        log.info(f'Docker import: Assuming container name is {import_input_path}. Correct?')
        return subprocess.run(cmd, check=True, shell=True)

    def process_docker_image_id(self, param_imageid=None):
        """Get and process the CWL compliant dockerImageId dockerRequirement.

        A divergence from the CWL spec. Docker image Id is defined by docker as a checksum
        on a container, not a human-readable name. The Docker image ID must be produced after
        the container is built, and can not be used to tag the container for that reason.
        The param_imageid may be used to override DockerRequirement specs.
        """
        # Parameter takes precedence
        if param_imageid:
            self.docker_image_id = param_imageid
        # If previously set by param, return 0 and ignore task.
        if self.docker_image_id:
            return 0

        # Need ImageId to know how dockerFile should be named, else fail
        task_image_id = self.get_docker_req('dockerImageId')

        # Set imageid
        self.docker_image_id = task_image_id

        # If task and parameter still doesn't specify ImageId, consider this an error.
        if not self.docker_image_id:
            return 0
        return 1

    def process_docker_output_directory(self, param_output_directory=None):
        """Get and process the CWL compliant dockerOutputDirectory dockerRequirement.

        CWL spec 09-23-2020: Set the designated output directory
        to a specific location inside the Docker container. The
        param_output_directory may be used to override DockerRequirement
        specs.
        """
        # Allow parameter over-ride.
        if param_output_directory:
            self.container_output_path = param_output_directory
        return 0

    def process_copy_container(self, force=False):
        """Get and process the BEE CWL extension copyContainer dockerRequirement.

        This CWL extension will copy an existing container to the build archive.

        If you have a container tarball, and all you need to do is stage it,
        that is, all you need to do is copy it to a location that BEE knows,
        use this to put the container into the build archive.
        """
        # Need container_path to know how dockerfile should be named, else fail
        task_container_path = self.get_docker_req('beeflow:copyContainer')
        if not task_container_path:
            log.error("beeflow:copyContainer: You must specify the path to an existing container.")
            return 1

        if self.container_name:
            copy_target = '/'.join([self.container_archive, self.container_name + '.tar.gz'])
        else:
            copy_target = '/'.join([self.container_archive,
                                    crt_driver.get_ccname(task_container_path) + '.tar.gz'])
        log.info(f'Build will copy a container to {copy_target}')
        # Return if image already exist and force==False.
        if os.path.exists(copy_target) and not force:
            log.info('Container by this name exists in archive. Taking no action.')
            return 0
        # Force remove any cached images if force==True
        if os.path.exists(copy_target) and force:
            try:
                os.remove(copy_target)
            except FileNotFoundError:
                pass

        log.info('Copying container.')
        cmd = (f'cp {task_container_path} {copy_target}\n'
               )
        log.info(f'Executing: {cmd}')
        return subprocess.run(cmd, check=True, shell=True)

    def process_container_name(self):
        """Get and process BEE CWL extension for containerName dockerRequirement.

        This is a BEE extension to CWL to refer to containers with human-readable name.

        The CWL spec currently uses dockerImageId to refer to the name of a container
        but this is explicitly not how Docker defines it. We need a way to name
        containers in a human readable format.
        """
        task_container_name = self.get_docker_req('beeflow:containerName')
        if not task_container_name and self.docker_image_id is None:
            log.error("beeflow:containerName: You must specify the containerName or dockerImageId")
            return 1
        self.container_name = task_container_name
        log.info(f'Setting container_name to: {self.container_name}')
        return 0


class SingularityBuildDriver(ContainerBuildDriver):
    """Driver interface between WFM and a container build system.

    A driver object must implement an __init__ method that
    requests a RTE, and a method to return the requested RTE.
    """

    def __init__(self, task, userconf_file=None):
        """Begin build request.

        Parse hints and requirements to determine target build
        system. Harvest relevant BeeConfig params in kwargs.

        :param requirements: the workflow requirements
        :type requirements: set of Requirement instances
        :param hints: the workflow hints (optional requirements)
        :type hints: set of Requirement instances
        :param kwargs: Dictionary of build system config options
        :type kwargs: set of build system parameters
        """

    def process_docker_pull(self, addr=None, force=False):
        """Get and process the CWL compliant dockerPull dockerRequirement.

        CWL spec 09-23-2020: Specify a Docker image to
        retrieve using docker pull. Can contain the immutable
        digest to ensure an exact container is used.
        """

    def process_docker_load(self):
        """Get and process the CWL compliant dockerLoad dockerRequirement.

        CWL spec 09-23-2020: Specify a HTTP URL from which to
        download a Docker image using docker load.
        """

    def process_docker_file(self, task_dockerfile=None, force=False):
        """Get and process the CWL compliant dockerFile dockerRequirement.

        CWL spec 09-23-2020: Supply the contents of a Dockerfile
        which will be built using docker build.
        """

    def process_docker_import(self, param_import=None):
        """Get and process the CWL compliant dockerImport dockerRequirement.

        CWL spec 09-23-2020: Provide HTTP URL to download and
        gunzip a Docker images using docker import. The param_import
        may be used to override DockerRequirement specs.
        """

    def process_docker_image_id(self, param_imageid=None):
        """Get and process the CWL compliant dockerImageId dockerRequirement.

        CWL spec 09-23-2020: The image id that will be used for
        docker run. May be a human-readable image name or the
        image identifier hash. May be skipped if dockerPull is
        specified, in which case the dockerPull image id must be
        used. The param_imageid may be used to override DockerRequirement
        specs.
        """

    def process_docker_output_directory(self, param_output_directory=None):
        """Get and process the CWL compliant dockerOutputDirectory.

        CWL spec 09-23-2020: Set the designated output directory
        to a specific location inside the Docker container. The
        param_output_directory may be used to override DockerRequirement
        specs.
        """
# Ignore W0231: linter doesn't know about abstract classes, it's ok to now call the parent __init__
# pylama:ignore=W0231



"""Abstract base class for crt_driver, the Container Runtime and drivers.

Builds text for job to run task in a Container
"""
from abc import ABC, abstractmethod


class ContainerRuntimeResult:
    """Result to be used for returning to the worker code."""

    def __init__(self, env_code, pre_commands, main_command, post_commands):
        """Construct the result."""
        self.env_code = env_code
        self.pre_commands = pre_commands
        self.main_command = main_command
        self.post_commands = post_commands


class ContainerRuntimeDriver(ABC):
    """ContainerRuntimeDriver interface for generic container runtime."""

    @abstractmethod
    def run_text(self, task):
        """Create commands for job using the container runtime.

        Returns a tuple (pre-commands, main-command, post-commands).
        :param task: instance of Task
        :rtype: tuple of (list of list of str, list of str, list of list of str)
        """

    @abstractmethod
    def build_text(self, userconfig, task):
        """Create text for builder pre-run using the container runtime.

        :param task: instance of Task
        :rtype: string
        """



"""Chameleon provider code."""
import openstack

from beeflow.common.cloud import provider


class ChameleoncloudProvider(provider.Provider):
    """Chameleoncloud provider class."""

    def __init__(self, stack_name=None, **_kwargs):
        """Chameleoncloud provider constructor."""
        self._stack_name = stack_name
        self._api = openstack.connect()

    def create_from_template(self, template_file):
        """Create from a template file."""
        raise RuntimeError(
            'create_from_template() is not implemented for Chameleoncloud. '
            'Use the Horizon interface instead'
        )

    def get_ext_ip_addr(self, node_name):  # noqa
        """Get the external IP address of the node, if it has one."""
        if self._stack_name is not None:
            stack = self._api.get_stack(self._stack_name)
            if stack is None:
                raise RuntimeError(f'Invalid stack {self._stack_name}')
            outputs = {output['output_key']: output['output_value'] for output in stack['outputs']}
            if 'head_node_login_ip' in outputs:
                return outputs['head_node_login_ip']
        return None



"""Configuration validor tests."""
import pytest
from beeflow.common.config_validator import ConfigValidator, ConfigError


def test_empty():
    """Test an empty config."""
    validator = ConfigValidator(description='empty test case')
    assert validator.validate({}) == {} # noqa
    # It should not allow any extra keys
    with pytest.raises(ValueError):
        validator.validate({'key': {}})


def test_two():
    """Test a config with two sections."""
    conf = {'section0': {'key0': 'value'}, 'section1': {'key0': '123'}}
    valid_conf = {'section0': {'key0': 'value'}, 'section1': {'key0': 123}}

    validator = ConfigValidator(description='test case of two')
    validator.section('section0', info='some info')
    validator.section('section1', info='some more info')
    validator.option('section0', 'key0', validator=str, info='some config opt')
    validator.option('section1', 'key0', validator=int, info='some config opt')

    assert validator.validate(conf) == valid_conf
    # Test missing sections and values
    with pytest.raises(ValueError):
        validator.validate({'section1': {'key0': 'abc'}})
    with pytest.raises(ValueError):
        validator.validate({})
    with pytest.raises(ValueError):
        validator.validate({'section2': {'key0': 'something'}})


def test_choices():
    """Test a config option that can have a set of choices."""
    validator = ConfigValidator(description='test case for choice-based option')
    validator.section('section0', info='info')
    validator.option('section0', 'choice-key', choices=('A', 'B', 'C'), info='choice-based option')

    assert (validator.validate({'section0': {'choice-key': 'B'}})
            == {'section0': {'choice-key': 'B'}}) # noqa
    with pytest.raises(ValueError):
        assert validator.validate({'section0': {'choice-key': 'E'}})


def test_depends_on():
    """Test the depends_on relation for sections in a config."""
    validator = ConfigValidator(description='depends on relation test')
    validator.section('one', info='section one')
    validator.option('one', 'key', choices=('A', 'B'), info='choice-based option')
    # Section two depends on one::key == A
    validator.section('two', info='section two', depends_on=('one', 'key', 'A'))
    validator.option('two', 'some-key', info='some other config value dependent on [one] key == A')

    with pytest.raises(ValueError):
        assert validator.validate({'one': {'key': 'B'}, 'two': {}})
    with pytest.raises(ValueError):
        validator.validate({'one': {'key': 'A'}})
    assert (validator.validate({'one': {'key': 'A'}, 'two': {'some-key': '123'}})
            == {'one': {'key': 'A'}, 'two': {'some-key': '123'}}) # noqa


def test_depends_on_order():
    """Test that the depends_on relation is not broken by ordering."""
    validator = ConfigValidator(description='depends on relation test')
    # Section two depends on one::key == A
    validator.section('two', info='section two', depends_on=('one', 'key', 'A'))
    validator.option('two', 'some-key', info='some other config value dependent on [one] key == A')
    validator.section('one', info='section one')
    validator.option('one', 'key', choices=('A', 'B'), info='choice-based option')

    # Sections must be listed in dependency order, so if section two dependends
    # on section one, then section one must be listed first, then section two
    sections = validator.sections
    assert sections[0][0] == 'one'
    assert sections[1][0] == 'two'


def test_depends_on_validate():
    """Test that validation does not break the depends_on relation."""
    validator = ConfigValidator(description='depends on relation test')
    # Section two depends on one::key == A
    validator.section('two', info='section two', depends_on=('one', 'key', 'A'))
    validator.option('two', 'some-key', info='some other config value dependent on [one] key == A')
    validator.section('one', info='section one')
    validator.option('one', 'key', choices=('A', 'B'), info='choice-based option')

    assert not validator.is_section_valid({'one': {'key': 'B'}}, 'two')
    assert validator.is_section_valid({'one': {'key': 'A'}}, 'two')
    assert validator.is_section_valid({}, 'one')


def test_section_order():
    """Ensure that the order of section insertion is preserved."""
    validator = ConfigValidator(description='section order test')
    validator.section('one', info='section one')
    validator.option('one', 'key-a', info='some option')
    validator.section('two', info='section two')
    validator.option('two', 'key-b', info='some option')

    sections = validator.sections

    assert sections[0][0] == 'one'
    assert sections[1][0] == 'two'


def test_double_definition():
    """Ensure that double definitions of sections and options are not allowed."""
    validator = ConfigValidator(
        description='test case for when code tries to define sections and options twice'
    )

    validator.section('section', info='some section')
    with pytest.raises(ConfigError):
        validator.section('section', info='some section defined twice')

    validator.option('section', 'option', info='some option')
    with pytest.raises(ConfigError):
        validator.option('section', 'option', info='some option defined twice')


def test_missing_section():
    """Test validation of a missing section."""
    validator = ConfigValidator(description='missing section')

    with pytest.raises(ConfigError):
        validator.option('some-section', 'option',
                         info='defining an option, but the section is not defined')


def test_option_attr():
    """Test adding attributes to a section option."""
    validator = ConfigValidator(description='option with attribute')

    validator.section('abc', info='some section')
    validator.option('abc', 'test', info='some option with attributes', attrs={'key': 1})

    opt_name, option = validator.options('abc')[0]
    assert opt_name == 'test'
    assert option.attrs == {'key': 1}



#!/usr/bin/env python3
"""bee-client.

This script provides an client interface to the user to manage workflows.
Capablities include submitting, starting, listing, pausing and cancelling workflows.
"""
import os
import sys
import logging
import inspect
import pathlib
import shutil
import subprocess
import textwrap
import time
import jsonpickle
import requests
import typer

from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.cli import NaturalOrderGroup
from beeflow.common.connection import Connection


# Length of a shortened workflow ID
short_id_len = 6 #noqa: Not a constant

# Maximum length of a workflow ID
MAX_ID_LEN = 32

# Global used to indicate whether this instance is interactive or not
_INTERACTIVE = False


logging.basicConfig(level=logging.WARNING)
WORKFLOW_MANAGER = 'bee_wfm/v1/jobs/'


class ClientError(Exception):
    """Client error class."""

    def __init__(self, *args):
        """Error constructor."""
        self.args = args


def error_exit(msg, include_caller=True):
    """Print a message and exit or raise an error with that message."""
    if include_caller:
        caller_func = str.capitalize(inspect.stack()[1].function)
        msg = caller_func + ': ' + msg
    if _INTERACTIVE:
        typer.secho(msg, fg=typer.colors.RED, file=sys.stderr)
        sys.exit(1)
    else:
        # Raise an error so that client libraries can handle it
        raise ClientError(msg) from None


def error_handler(resp): # noqa (this is an error handler, it doesn't need to return an expression)
    """Handle a 500 error in a response."""
    if resp.status_code != 500:
        return resp
    data = resp.json()
    if 'error' not in data:
        return resp
    error = data['error']
    error_file = f'bee-error-{int(time.time())}.log'
    # Save the args and exception info to the file
    with open(error_file, 'w', encoding='utf-8') as fp:
        fp.write(f'args: {" ".join(sys.argv)}\n')
        fp.write(error)
    msg = ('An error occurred with the WFM. Please save your workflow and the '
           f'error log "{error_file}". If possible, please report this to the '
           'BEE development team.')
    msg = textwrap.fill(msg)
    error_exit(msg, include_caller=False)


def _wfm_conn():
    """Return a connection to the WFM."""
    return Connection(bc.get('workflow_manager', 'socket'),
                      error_handler=error_handler)


def _url():
    #    """Returns URL to the workflow manager end point"""
    #    wfm_listen_port = wf_db.get_wfm_port()
    #    workflow_manager = 'bee_wfm/v1/jobs'
    #    return f'http://127.0.0.1:{wfm_listen_port}/{workflow_manager}/'
    """Return URL to the workflow manager end point."""
    return WORKFLOW_MANAGER


def _short_id(wf_id):
    return wf_id[:short_id_len]


def _resource(tag=""):
    return _url() + str(tag)


def check_short_id_collision():
    """Check short workflow IDs for colliions; increase short ID length if detected."""
    global short_id_len  #noqa: Not a constant
    conn = _wfm_conn()
    resp = conn.get(_url(), timeout=60)
    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit(f"Checking for ID collision failed: {resp.status_code}")

    workflow_list = jsonpickle.decode(resp.json()['workflow_list'])
    if workflow_list:
        while short_id_len < MAX_ID_LEN:
            id_list = [_short_id(job[1]) for job in workflow_list]
            id_list_set = set(id_list)
            # Collision if set shorter than list
            if len(id_list_set) < len(id_list):
                short_id_len += 1
            else:
                break
        else:
            raise RuntimeError("collision detected between two full workflow IDs")
    else:
        print("There are currently no jobs.")


def match_short_id(wf_id):
    """Match user-provided short workflow ID to full workflow IDs."""
    matched_ids = []

    try:
        conn = _wfm_conn()
        resp = conn.get(_url(), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit(f'Could not match ID: {wf_id}. Code {resp.status_code}')
        # raise ApiError("GET /jobs".format(resp.status_code))

    workflow_list = jsonpickle.decode(resp.json()['workflow_list'])
    if workflow_list:
        for job in workflow_list:
            if job[1].startswith(wf_id):
                matched_ids.append(job[1])
        if len(matched_ids) > 1:
            logging.info(f"user-provided workflow ID {wf_id} matched multiple stored workflow IDs")
            error_exit("provided workflow ID ambiguous")
        elif not matched_ids:
            logging.info(f"user-provided workflow ID {wf_id} did not match any"
                         "stored workflow ID")
            error_exit("Provided workflow ID does not match any submitted "
                       "workflows")
        else:
            logging.info(f"user-provided workflow ID {wf_id} matched stored"
                         "workflow ID {matched_ids[0]}")
            long_wf_id = matched_ids[0]
            return long_wf_id
    else:
        print("There are currently no workflows.")

    return None


app = typer.Typer(no_args_is_help=True, add_completion=False, cls=NaturalOrderGroup)


@app.command()
def submit(wf_name: str = typer.Argument(..., help='The workflow name'),
           wf_path: pathlib.Path = typer.Argument(..., help='Path to the workflow .tgz or dir'),
           main_cwl: str = typer.Argument(..., help='filename of main CWL file'),
           yaml: str = typer.Argument(..., help='filename of YAML file'),
           workdir: pathlib.Path = typer.Argument(...,
           help='working directory for workflow containing input + output files',)):
    """Submit a new workflow."""
    tarball_path = ""
    if os.path.exists(wf_path):
        # Check to see if the wf_path is a tarball or a directory. Run package() if directory
        if os.path.isdir(wf_path):
            print("Detected directory instead of packaged workflow. Packaging Directory...")
            bee_workdir = bc.get('DEFAULT', 'bee_workdir')
            package(wf_path, pathlib.Path(bee_workdir))
            tarball_path = pathlib.Path(bee_workdir + "/" + str(wf_path.name) + ".tgz")
            wf_tarball = open(tarball_path, 'rb')
        else:
            wf_tarball = open(wf_path, 'rb')
    else:
        error_exit(f'Workflow tarball {wf_path} cannot be found')

    # Make sure the workdir is an absolute path and exists
    workdir = os.path.expanduser(workdir)
    workdir = os.path.abspath(workdir)
    if not os.path.exists(workdir):
        error_exit(f"Workflow working directory \"{workdir}\" doesn't exist")

    # TODO: Can all of this information be sent as a file?
    data = {
        'wf_name': wf_name.encode(),
        'wf_filename': os.path.basename(wf_path).encode(),
        'main_cwl': main_cwl,
        'yaml': yaml,
        'workdir': workdir
    }
    files = {
        'workflow_archive': wf_tarball
    }
    try:
        conn = _wfm_conn()
        resp = conn.post(_url(), data=data, files=files, timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.created:  # pylint: disable=no-member
        if resp.status_code == 400:
            data = resp.json()
            error_exit(data['msg'])
        error_exit(f"Submit for {wf_name} failed. Please check the WF Manager.")

    check_short_id_collision()
    if 'wf_id' not in resp.json():
        error_exit("wf_id not in WFM response")
    wf_id = resp.json()['wf_id']
    typer.secho("Workflow submitted! Your workflow id is "
                f"{_short_id(wf_id)}.", fg=typer.colors.GREEN)
    logging.info('Sumit workflow:  {resp.text}')

    # Cleanup code
    if tarball_path:
        os.remove(tarball_path)
    return wf_id


@app.command()
def start(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Start a workflow with a workflow ID."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.post(_resource(long_wf_id), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit(f"Starting {long_wf_id} failed."
                   f" Returned {resp.status_code}")

    typer.echo(f"{resp.json()['msg']}")
    logging.info('Started  {resp.text}')


@app.command()
def package(wf_path: pathlib.Path = typer.Argument(...,
            help='Path to the workflow package directory'),
            package_dest: pathlib.Path = typer.Argument(...,
            help='Path for where the packaged workflow should be saved')
            ):
    """Package a workflow into a tarball."""
    if not os.path.isdir(wf_path):
        error_exit(f"{wf_path} is not a valid directory.")

    # Get the filename
    wf_dir = wf_path.name
    tarball = wf_path.name + '.tgz'
    # Get the parent directory and resolve it in case the path is relative
    parent_dir = wf_path.resolve().parent
    # Just use tar with subprocess. Python's tar library is not performant.
    return_code = subprocess.run(['tar', '-C', parent_dir, '-czf', tarball, wf_dir],
                                 check=True).returncode
    package_path = package_dest.resolve()/tarball  # noqa: Not an arithmetic operation

    # Get the curent working directory
    cwd = pathlib.Path().absolute()
    if package_dest != cwd:
        # Move the tarball if the directory it's wanted in is not in the current working directory
        tarball_path = cwd/tarball  # noqa: Not an arithmetic operation
        shutil.move(tarball_path, package_path)

    if return_code != 0:
        error_exit("Package failed")
    else:
        print(f"Package {tarball} created successfully")


@app.command()
def listall():
    """List all worklfows."""
    try:
        conn = _wfm_conn()
        resp = conn.get(_url(), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager did not return workflow list')

    logging.info('List Jobs:  {resp.text}')
    workflow_list = jsonpickle.decode(resp.json()['workflow_list'])
    if workflow_list:
        typer.secho("Name\tID\tStatus", fg=typer.colors.GREEN)

        for name, wf_id, status in workflow_list:
            typer.echo(f"{name}\t{_short_id(wf_id)}\t{status}")
    else:
        typer.echo("There are currently no workflows.")

    logging.info('List workflows:  {resp.text}')


@app.command()
def query(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Get the status of a workflow."""
    # wf_id is a tuple with the short version and long version
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.get(_resource(long_wf_id), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('Could sucessfully query workflow manager')

    tasks_status = resp.json()['tasks_status']
    wf_status = resp.json()['wf_status']
    if tasks_status == 'Unavailable':
        typer.echo(wf_status)
    else:
        typer.echo(wf_status)
        typer.echo(tasks_status)

    logging.info('Query workflow:  {resp.text}')
    return wf_status, tasks_status


@app.command()
def pause(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Pause a workflow (Running tasks will finish)."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.patch(_resource(long_wf_id), json={'option': 'pause'},
                          timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')
    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager could not pause workflow.')
    logging.info('Pause workflow:  {resp.text}')


@app.command()
def resume(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Resume a paused workflow."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.patch(_resource(long_wf_id),
                          json={'wf_id': long_wf_id, 'option': 'resume'},
                          timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')
    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager could not resume workflow.')
    logging.info('Resume workflow:  {resp.text}')


@app.command()
def cancel(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Cancel a workflow."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.delete(_resource(long_wf_id), timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')
    if resp.status_code != requests.codes.accepted:  # pylint: disable=no-member
        error_exit('WF Manager could not cancel workflow.')
    typer.secho("Workflow cancelled!", fg=typer.colors.GREEN)
    logging.info(f'Cancel workflow: {resp.text}')


@app.command()
def copy(wf_id: str = typer.Argument(..., callback=match_short_id)):
    """Copy an archived workflow."""
    long_wf_id = wf_id
    try:
        conn = _wfm_conn()
        resp = conn.patch(_url(), files={'wf_id': long_wf_id}, timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')
    if resp.status_code != requests.codes.okay:  # pylint: disable=no-member
        error_exit('WF Manager could not copy workflow.')
    archive_file = jsonpickle.decode(resp.json()['archive_file'])
    archive_filename = resp.json()['archive_filename']
    logging.info(f'Copy workflow: {resp.text}')
    return archive_file, archive_filename


@app.command()
def reexecute(wf_name: str = typer.Argument(..., help='The workflow name'),
              wf_path: pathlib.Path = typer.Argument(..., help='Path to the workflow archive'),
              workdir: pathlib.Path = typer.Argument(
                  ...,
                  help='working directory for workflow containing input + output files')
              ):
    """Reexecute an archived workflow."""
    if os.path.exists(wf_path):
        wf_tarball = open(wf_path, 'rb')
    else:
        error_exit(f'Workflow tarball {wf_path} cannot be found')

    data = {
        'wf_filename': os.path.basename(wf_path).encode(),
        'wf_name': wf_name.encode(),
        'workdir': workdir
    }
    try:
        conn = _wfm_conn()
        resp = conn.put(_url(), data=data,
                        files={'workflow_archive': wf_tarball}, timeout=60)
    except requests.exceptions.ConnectionError:
        error_exit('Could not reach WF Manager.')

    if resp.status_code != requests.codes.created: #noqa: member does exist
        error_exit(f"Reexecute for {wf_name} failed. Please check the WF Manager.")

    wf_id = resp.json()['wf_id']
    typer.secho("Workflow submitted! Your workflow id is "
                f"{_short_id(wf_id)}.", fg=typer.colors.GREEN)
    logging.info(f'ReExecute Workflow: {resp.text}')
    return wf_id


def main():
    """Execute bee_client."""
    global _INTERACTIVE
    _INTERACTIVE = True
    bc.init()
    app()


if __name__ == "__main__":
    app()

# Pylint is reporting no member for requests.codes even when they exist
#     ignoring them line by line
# Ignore using with for open files; used to send command.
# Ignore W0511: This is a TODO that should be addressed later
# pylama:ignore=R1732,W0511



"""BEE Cloud class."""


class CloudError(Exception):
    """Cloud error class."""

    def __init__(self, msg):
        """Cloud error constructor."""
        self.msg = msg
# Ignore W0231: This is a user-defined exception so the base class doesn't
#               usually need to be called.
# pylama:ignore=W0231



"""Code implementing various scheduling algorithms.

Code implementing scheduling algorithms, such as FCFS, Backfill, etc.
"""

import abc
import os
import time

from beeflow.scheduler import resource_allocation


class Algorithm(abc.ABC):
    """Scheduling algorithm abstract class.

    Base abstract class for implementing a scheduling algorithm.
    """

    @staticmethod
    @abc.abstractmethod
    def load(**kwargs):
        """Load configuration for the algorithm."""

    @staticmethod
    @abc.abstractmethod
    def schedule_all(tasks, resources, **kwargs):
        """Schedule all tasks with the implemented algorithm.

        Schedule all tasks with the implemented algorithm.
        :param tasks: list of tasks to schedule
        :type tasks: list of instance of Task
        :param resources: list of resources
        :type resources: list of instance of resource_allocation.Resource
        """


class SJF(Algorithm):
    """Shortest job first algorithm.

    Class holding scheduling code for runing the shortest job first algorithm.
    """

    @staticmethod
    def load(**kwargs):
        """Load algorithm configuration, if necessary."""

    @staticmethod
    def schedule_all(tasks, resources, **kwargs):
        """Schedule a list of independent tasks with SJF.

        Schedule a list of independent tasks with SFJ.
        :param tasks: list of tasks to schedule
        :type tasks: list of instance of Task
        :param resources: list of resources
        :type resources list of instance of resource_allocation.Resource
        """
        # First sort the tasks by how long they are, then send them off to
        # FCFS
        tasks = tasks[:]
        tasks.sort(key=lambda task: task.requirements.max_runtime)
        FCFS.schedule_all(tasks, resources, **kwargs)


class FCFS(Algorithm):
    """FCFS scheduling algorithm."""

    @staticmethod
    def load(**kwargs):
        """Load algorithm configuration, if necessary."""

    @staticmethod
    def schedule_all(tasks, resources, **kwargs):
        """Schedule a list of independent tasks with FCFS.

        Schedule an entire list of tasks using FCFS. Tasks that
        cannot be allocated will be left with an empty allocations
        property.
        :param tasks: list of tasks to schedule
        :type tasks: list of instance of Task
        :param resources: list of resources
        :type resources: list of instance of resource_allocation.Resource
        """
        allocator = resource_allocation.TaskAllocator(resources)
        start_time = 0
        for task in tasks:
            if not allocator.fits_requirements(task.requirements):
                continue
            # Find the start_time
            while not allocator.can_run_now(task.requirements, start_time):
                start_time = allocator.get_next_end_time(start_time)
            task.allocations = allocator.allocate(task.requirements,
                                                  start_time)


class Backfill(Algorithm):
    """Backfill scheduling algorithm.

    This class holds the scheduling code used for the Backfill
    scheduling algorithm.
    """

    @staticmethod
    def load(**kwargs):
        """Load algorithm configuration, if necessary."""

    @staticmethod
    def schedule_all(tasks, resources, **kwargs):
        """Schedule a list of independent tasks with Backfill.

        Schedule an entire list of tasks using Backfill. Tasks that
        cannot be allocated will be left with an empty allocations
        property.
        :param tasks: list of tasks to schedule
        :type tasks: list of instance of Task
        :param resources: list of resources
        :type resources: list of instance of resource_allocation.Resource
        """
        tasks = tasks[:]
        current_time = 0
        allocator = resource_allocation.TaskAllocator(resources)
        while tasks:
            task = tasks.pop(0)
            # Can this task run at all?
            if not allocator.fits_requirements(task.requirements):
                continue
            # Can this task run immediately?
            start_time = current_time
            # max_runtime = task.requirements.max_runtime
            if allocator.can_run_now(task.requirements, start_time):
                allocs = allocator.allocate(task.requirements, start_time)
                task.allocations = allocs
                continue
            # This job must run later, so we need to find the shadow time
            # (earliest time at which the job can run)
            times = allocator.get_end_times()
            times.sort()
            shadow_time = 0
            for shadow_time in times:
                if allocator.can_run_now(task.requirements, shadow_time):
                    allocs = allocator.allocate(task.requirements, shadow_time)
                    task.allocations = allocs
                    break
            # Now backfill other tasks
            times.insert(0, current_time)
            remaining = []
            for backfill_task in tasks:
                max_runtime = backfill_task.requirements.max_runtime
                possible_times = [start_time for start_time in times
                                  if (start_time + max_runtime) < shadow_time]
                for start_time in possible_times:
                    if allocator.can_run_now(backfill_task.requirements,
                                             start_time):
                        allocs = allocator.allocate(backfill_task.requirements,
                                                    start_time)
                        backfill_task.allocations = allocs
                # Could not backfill this task
                if not backfill_task.allocations:
                    remaining.append(backfill_task)
            # Tasks in remaining cannot be backfilled and must be run later
            # Reset the tasks to the remaining list
            tasks = remaining


class AlgorithmLogWrapper:
    """Algorithm log wrapper class.

    Algorithm wrapper class to be used as a wrap to log the task scheduling
    data for future training and other extra information.
    """

    def __init__(self, cls, alloc_logfile='schedule_log.txt', **kwargs):
        """Algorithm log wrapper class constructor.

        Algorithm wrapper class constructor.
        :param cls: object to pass operations onto
        :type cls: algorithm object
        :param alloc_logfile: name of logfile to write task scheduling to
        :type alloc_logfile: str
        :param kwargs: key word arguments to pass to schedule_all()
        :type kwargs: instance of dict
        """
        self.cls = cls
        self.alloc_logfile = alloc_logfile
        self.kwargs = kwargs

    def schedule_all(self, tasks, resources):
        """Schedule all tasks using the internal class and log results.

        Schedule all of the tasks with the internal class and write the
        results out to a log file.
        """
        self.cls.schedule_all(tasks, resources, **self.kwargs)
        # Make the directory, just in case it doesn't exist already
        os.makedirs(os.path.dirname(self.alloc_logfile), exist_ok=True)
        with open(self.alloc_logfile, 'a', encoding='utf-8') as fp:
            print('; Log start at', time.time(), file=fp)
            # curr_allocs = []
            for task in tasks:
                # TODO: Rethink this log output

                # possible_allocs = build_allocation_list(task, tasks,
                #                                         resources,
                #                                         curr_allocs)
                # Find the value of a - the index of the allocation for this
                # task
                # a = -1
                # TODO: Calculation of a needs to change
                # if task.allocations:
                #     start_time = task.allocations[0].start_time
                #     # a should be the first alloc with the same start_time
                #     for i, alloc in enumerate(possible_allocs):
                #         if alloc[0].start_time == start_time:
                #             a = i
                #             break
                # Output in SWF format
                # TODO: These variables may not be all in the right spot and
                # some may be missing as well
                print(-1, -1, -1, task.requirements.max_runtime,
                      task.requirements.nodes, task.requirements.max_runtime,
                      task.requirements.mem_per_node, task.requirements.nodes,
                      -1, task.requirements.mem_per_node,
                      task.requirements.mem_per_node, -1,
                      # task.requirements.mem, task.requirements.nodes, -1,
                      # task.requirements.mem, task.requirements.mem, -1,
                      -1, -1, -1, -1, -1, -1, -1, file=fp)
                # print(*vec, file=fp)
                # curr_allocs.extend(task.allocations)


# TODO: Perhaps this value should be a config value
MEDIAN = 2


algorithm_objects = {
    'fcfs': FCFS,
    'backfill': Backfill,
    'sjf': SJF,
}


def load(algorithm=None, **kwargs):  # noqa ('algorithm' may be used in the future)
    """Load data needed by the algorithms.

    Load data needed by algorithms, if necessary.
    """
    FCFS.load(**kwargs)
    Backfill.load(**kwargs)
    SJF.load(**kwargs)


def choose(algorithm=None, default_algorithm=None, **kwargs):
    """Choose which algorithm to run at this point.

    Determine which algorithm class needs to run and return it.
    :param tasks: list of tasks:
    :type tasks: list of instance of Task
    :rtype: class derived from Algorithm (not an instance)
    """
    # Choose the default algorithm
    default = default_algorithm if default_algorithm is not None else 'fcfs'
    cls = algorithm_objects[default]
    if algorithm is not None:
        cls = algorithm_objects[algorithm]
    return AlgorithmLogWrapper(cls, **kwargs)
# Ignoring E0211: This is how the class is designed right now, we should think about changing this
#                 however.
# Ignoring W0511: A number of these TODO's are hinted at in issue #333, but I don't want to remove
#                 them from the code until this issue is fully addressed.
# Ignoring R0903: Too few public methods, not sure how this calculated and this will be fixed with
#                 issue #333
# pylama:ignore=E0211,C0415,W0511,R0903



"""Task allocator code."""
from beeflow.scheduler import serializable


class TaskAllocator:
    """Task allocator class.

    Class used to help manage allocation of tasks onto multiple resources.
    """

    def __init__(self, resources):
        """Task allocator constructor.

        :param resources: available resources that can be allocated
        :type resources: list of instance of Resource
        """
        self.allocations = []
        self.resources = resources

    def _fits_requirements_with_overlap(self, reqs, overlap):
        """Check if a task with these requirements can run with overlap.

        This determines if a task can run given `overlap`, a list of
        overlapping allocations.
        :param reqs: task requirements
        :type reqs: instance of Requirements
        :param overlap: overlappinig allocations
        :type overlap: list of instance of Allocation
        :rtype: bool
        """
        total_nodes = 0
        # Count the total number of nodes that match the required properties
        for res in self.resources:
            # TODO: Handle shared nodes (perhaps with a shared option)
            allocs = [alloc for alloc in overlap if alloc.id_ == res.id_]
            # if (reqs.mem_per_node <= res.mem_per_node
            #     and reqs.gpus_per_node <= res.gpus_per_node):
            if res.fits(reqs):
                total_nodes += (res.nodes - sum(alloc.nodes for alloc in allocs))
        # Return True if we have enough nodes that match
        return total_nodes >= reqs.nodes

    def _calculate_overlap(self, start_time, max_runtime):
        """Calculate the overlap for the time period [start_time, max_runtime].

        :param start_time: start time of the period
        :type start_time: int
        :param max_runtime: max runtime for the time period
        :type max_runtime: int
        :rtype: list of instance of Allocation
        """
        # TODO: This calculation be off
        return [alloc for alloc in self.allocations
                if (start_time < (alloc.start_time + alloc.max_runtime)
                    and (start_time + max_runtime) > alloc.start_time)]

    def fits_requirements(self, reqs):
        """Determine if the resources can fit the requirements given.

        :param reqs: requirements
        :type reqs: instance of Requirements
        :rtype: bool
        """
        return self._fits_requirements_with_overlap(reqs, [])

    def can_run_now(self, reqs, start_time):
        """Determine if a task with the requirements can run at start_time.

        :param reqs: requirements of a task
        :type reqs: instance of Requirements
        :param start_time: start time of a task
        :type start_time: int
        :rtype: bool
        """
        # Calculate the overlapping allocations
        overlap = self._calculate_overlap(start_time, reqs.max_runtime)
        return self._fits_requirements_with_overlap(reqs, overlap)

    def get_next_end_time(self, start_time):
        """Return the next finish time after start_time.

        This returns the next finish time of the current allocations stored in
        the allocator. If no finish time after this start_time exists then
        start_time will just be returned.
        :param start_time: start time of a possible task
        :type start_time: int
        """
        end_times = self.get_end_times()
        for end_time in end_times:
            if end_time > start_time:
                return end_time
        return start_time

    def get_end_times(self):
        """Get a list of ending times for all allocations.

        :rtype: list of int
        """
        # End times should be unique
        return list(set(a.start_time + a.max_runtime
                        for a in self.allocations))

    def allocate(self, reqs, start_time):
        """Allocate some allocations meeting the requirements at start_time.

        Note: This assumes that this allocation can be made and that there are
        no other overlapping allocations that could conflict.
        :param reqs: task requirements
        :type reqs: instance of Requirements
        :param start_time: start time of the task allocation
        :type start_time: int
        :rtype: list of instance of Allocation
        """
        # TODO
        overlap = self._calculate_overlap(start_time, reqs.max_runtime)
        allocs = []
        # Total number of nodes already allocated
        total_nodes = 0
        for res in self.resources:
            if total_nodes >= reqs.nodes:
                # Stop when we have enough nodes scheduled
                break
            other_allocs = [alloc for alloc in overlap if alloc.id_ == res.id_]
            # TODO: Replace this with a method res.fits()
            # if (reqs.mem_per_node <= res.mem_per_node
            #     and reqs.gpus_per_node <= res.gpus_per_node):
            if res.fits(reqs):
                used_nodes = sum(alloc.nodes for alloc in other_allocs)
                avail_nodes = res.nodes - used_nodes
                if avail_nodes > 0:
                    nodes = reqs.nodes - total_nodes
                    nodes = nodes if nodes < avail_nodes else avail_nodes
                    total_nodes += nodes
                    # Allocations should just contain a reference to the
                    # resource, by id_, rather than duplicating all the
                    # properties
                    alloc = Allocation(id_=res.id_, start_time=start_time,
                                       max_runtime=reqs.max_runtime,
                                       nodes=nodes)
                    allocs.append(alloc)
        # Add the new allocations to the stored allocations
        self.allocations.extend(allocs)
        return allocs


class Resource(serializable.Serializable):
    """Resource class.

    Resource class representing a resource.
    """

    def __init__(self, id_, nodes=1, mem_per_node=8192, gpus_per_node=0):
        """Resource class constructor.

        :param id_: ID of the resource
        :type id_: str
        :param nodes: number of nodes
        :type nodes: int
        :param mem_per_node: amount of memory
        :type mem_per_node: int
        :param gpus_per_node: number of GPUs per node
        :type gpus_per_node: int
        """
        self.id_ = id_
        self.nodes = nodes
        self.mem_per_node = mem_per_node
        self.gpus_per_node = gpus_per_node

    def fits(self, reqs):
        """Return True if the requirements fit this resource.

        :param reqs: task requirements
        :type reqs: instance of Requirements
        :rtype: bool
        """
        return (reqs.mem_per_node <= self.mem_per_node
                and reqs.gpus_per_node <= self.gpus_per_node)

    @staticmethod
    def decode(data):
        """Decode the resource.

        :param data: data representing the resource
        :type data: dict
        """
        return Resource(**data)


class Requirements(serializable.Serializable):
    """Requirements class.

    Requirements class representing task requirements.
    """

    # TODO: Determine default requirements
    def __init__(self, max_runtime, nodes=1, mem_per_node=1024,
                 gpus_per_node=0, cost=1):
        """Construct a requirements object.

        :param max_runtime: maximum runtime in seconds
        :type max_runtime: int
        :param nodes: number of nodes in total
        :type nodes: int
        :param mem_per_node: amount of memory per node
        :type mem_per_node: int
        :param gpus_per_node: number of gpus per node
        :type gpus_per_node: int
        :param cost: cost value
        :type cost: float
        """
        self.max_runtime = max_runtime
        self.nodes = nodes
        self.mem_per_node = mem_per_node
        self.gpus_per_node = gpus_per_node
        # TODO: Determine what other requirement properties are needed
        # TODO: Determine what the cost should be and how it is computed
        self.cost = cost

    @staticmethod
    def decode(data):
        """Decode the requirements.

        :param data: data representing the requirements
        :type data: dict
        """
        return Requirements(**data)


class Allocation(serializable.Serializable):
    """Allocation class.

    This represents an allocation for a task on a single resource.
    """

    def __init__(self, id_, start_time, max_runtime, nodes):
        """Allocation constructor.

        :param id_: ID of the resource
        :type id_: str
        :param start_time: start time of the allocation (seconds)
        :type start_time: int
        :param max_runtime: maximum runtime of the allocation (seconds)
        :type max_runtime: int
        :param nodes: number of nodes allocated
        :type nodes: int
        """
        self.id_ = id_
        self.start_time = start_time
        self.max_runtime = max_runtime
        # TODO: Determine what other allocation properties are needed (other
        # than just nodes)
        self.nodes = nodes

    @staticmethod
    def decode(data):
        """Decode the allocation.

        :param data: data representing the allocation
        :type data: dict
        """
        return Allocation(**data)
# Ignore W0511: This is related to issue #333.
# pylama:ignore=W0511



"""Test the resource_allocation submodule of BEE."""
import json

from beeflow.scheduler import resource_allocation


def test_resource_simple():
    """Test creating a resource."""
    resource = resource_allocation.Resource(id_='test-resource', nodes=1)

    assert resource.id_ == 'test-resource'
    assert resource.nodes == 1
    assert resource.mem_per_node == 8192
    assert resource.gpus_per_node == 0


def test_resource_encode_decode_json():
    """Ensure that a resource can be encoded/decoded to/from JSON."""
    resource = resource_allocation.Resource(id_='test-resource', nodes=1)

    s = json.dumps(resource.encode())

    decoded = resource_allocation.Resource.decode(json.loads(s))
    assert decoded.id_ == 'test-resource'
    assert decoded.nodes == resource.nodes
    assert decoded.mem_per_node == resource.mem_per_node
    assert decoded.gpus_per_node == resource.gpus_per_node


def test_allocation_simple():
    """Test creating an allocation."""
    allocation = resource_allocation.Allocation(id_='resource-0', start_time=0,
                                                max_runtime=4, nodes=2)

    assert allocation.id_ == 'resource-0'
    assert allocation.start_time == 0
    assert allocation.max_runtime == 4
    assert allocation.nodes == 2


def test_allocation_encode_decode_json():
    """Test that an allocation can be encoded/decoded to/from JSON."""
    allocation = resource_allocation.Allocation(id_='resource-0', start_time=3,
                                                max_runtime=5, nodes=6)

    s = json.dumps(allocation.encode())

    decoded = resource_allocation.Allocation.decode(json.loads(s))
    assert decoded.id_ == allocation.id_
    assert decoded.start_time == allocation.start_time
    assert decoded.max_runtime == allocation.max_runtime
    assert decoded.nodes == allocation.nodes


def test_requirements_simple():
    """Test creating a requirement."""
    requirements = resource_allocation.Requirements(max_runtime=3, nodes=1)

    assert requirements.max_runtime == 3
    assert requirements.nodes == 1
    # TODO: Determine what the default per node requirements should be


def test_requirements_encode_decode_json():
    """Test that an allocation can be encoded/decoded to/from JSON."""
    requirements = resource_allocation.Requirements(max_runtime=1, nodes=3)

    s = json.dumps(requirements.encode())

    decoded = resource_allocation.Requirements.decode(json.loads(s))
    assert decoded.max_runtime == requirements.max_runtime
    assert decoded.nodes == requirements.nodes
# Ignore W0511: This is related to issue #333
# pylama:ignore=W0511



"""Scheduler database code."""
from contextlib import contextmanager
import sqlite3
import jsonpickle


class Resources:
    """Resources wrapper class."""

    def __init__(self, conn):
        """Create the resources class."""
        self._conn = conn

    def __iter__(self):
        """Create an iterator over the resources."""
        cur = self._conn.cursor()
        res = cur.execute('SELECT id, resource FROM resources')
        while True:
            item = res.fetchone()
            if item is None:
                break
            resource = jsonpickle.decode(item['resource'])
            yield resource

    def extend(self, resources):
        """Add a list of new resources."""
        cur = self._conn.cursor()
        for resource in resources:
            data = jsonpickle.encode(resource)
            cur.execute('INSERT INTO resources (resource) VALUES (:resource)', {'resource': data})
        self._conn.commit()

    def clear(self):
        """Remove all resources."""
        cur = self._conn.cursor()
        cur.execute('DELETE FROM resources')
        self._conn.commit()


class SchedDB:
    """Scheduler database."""

    def __init__(self, conn):
        """Construct a new scheduler database connection."""
        self._conn = conn
        self._init_tables()

    def _init_tables(self):
        """Initialze the scheduler tables if they don't exist."""
        script = """CREATE TABLE IF NOT EXISTS resources(
                        id INTEGER PRIMARY KEY ASC,
                        resource TEXT);"""
        self._conn.executescript(script)

    @property
    def resources(self):
        """Get resources from the database."""
        return Resources(self._conn)


@contextmanager
def open_db(fname):
    """Open and return a new database."""
    with sqlite3.connect(fname) as conn:
        conn.row_factory = sqlite3.Row
        yield SchedDB(conn)
